from cli.utils import download_url, fetch_auth, parse_filter
from zpy.files import to_pathlib_path
import json
import requests

DATASET_TYPES = ["uploaded-data-sets", "generated-data-sets", "job-data-sets"]


@fetch_auth
def filter_datasets(dfilter, url, auth_headers):
    """filter datasets

    Filter dataset objects on ZumoLabs backend by given dfilter.
    Parse dfilter using parse_filter and make calls to
    generated, uploaded, job dataset endpoints.

    Args:
        dfilter (str): filter query for datasets
        url (str): backend endpoint
        auth_headers: authentication for backend

    Return:
        dict: filtered datasets by dfilter {'name': 'id'}
    """
    filtered_datasets = {}
    field, pattern, regex = parse_filter(dfilter)
    for dataset_type in DATASET_TYPES:
        endpoint = f"{url}/api/v1/{dataset_type}/?{field}__{pattern}={regex}"
        while endpoint is not None:
            r = requests.get(endpoint, headers=auth_headers)
            if r.status_code != 200:
                r.raise_for_status()
            response = json.loads(r.text)
            for r in response["results"]:
                filtered_datasets[r["name"]] = r["id"]
            endpoint = response["next"]
    return filtered_datasets


@fetch_auth
def create_generated_dataset(name, sim_name, config, url, auth_headers):
    """create dataset

    Create generated dataset on ZumoLabs backend which will launch
    a generation job with specified params.

    Args:
        name (str): name of dataset
        sim_name (str): name of sim the dataset is built from
        config (dict): configration of sim for this dataset
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/sims/"
    r = requests.get(endpoint, params={"name": sim_name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response["count"] != 1:
        raise NameError(f"found {response['count']} sims for name {sim_name}")
    sim = response["results"][0]
    endpoint = f"{url}/api/v1/generated-data-sets/"
    r = requests.post(
        endpoint,
        data={"sim": sim["id"], "config": json.dumps(config), "name": name},
        headers=auth_headers,
    )
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def create_uploaded_dataset(name, path, url, auth_headers):
    """upload dataset

    Upload dataset to S3 through ZumoLabs backend and
    create object.

    Args:
        name (str): name of dataset
        path (str): path to dataset
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/uploaded-data-sets/"
    r = requests.post(
        endpoint,
        data={"name": name},
        files={"file": open(path, "rb")},
        headers=auth_headers,
    )
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def download_dataset(name, path, dataset_type, url, auth_headers):
    """download dataset

    Download dataset object from S3 through ZumoLabs backend.

    Args:
        name (str): name of dataset to download
        path (str): output directory
        dataset_type (str): type of dataset to download
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        str: output file path
    """
    endpoint = f"{url}/api/v1/{dataset_type}-data-sets/"
    r = requests.get(endpoint, params={"name": name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response["count"] != 1:
        raise NameError(f"found {response['count']} datasets for name {name}")
    dataset = response["results"][0]
    endpoint = (
        f"{url}/api/v1/{dataset['dataset_type']}-data-sets/{dataset['id']}/download"
    )
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    name_slug = f"{dataset['name'].replace(' ', '_')}-{dataset['id'][:8]}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(response["redirect_link"], output_path)
    return output_path


def fetch_datasets():
    """fetch datasets

    Fetch dataset objects from ZumoLabs backend. Fetches all three
    datasets types uploaded, generated, and job.

    Returns:
        list: paginated sorted datasets for all types
    """
    datasets = []
    for dataset_type in DATASET_TYPES:
        short_dataset_type = dataset_type.split("-")[0]
        datasets += [
            (lambda d: d.update({"type": short_dataset_type}) or d)(d)
            for d in _fetch_type_datasets(dataset_type)
        ]
    return sorted(datasets, key=lambda i: i["created_at"], reverse=True)


@fetch_auth
def _fetch_type_datasets(dataset_type, url, auth_headers):
    """fetch type of datasets

    Args:
        dataset_type (str): type of dataset to fetch
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        list: paginated datasets for given type
    """
    endpoint = f"{url}/api/v1/{dataset_type}/"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]



from cli.loader import Loader
from cli.utils import parse_args, resolve_sweep
from cli.config import initialize_config, read_config, write_config, get_endpoint
from requests.auth import HTTPBasicAuth
from table_logger import TableLogger
from zpy.files import read_json, to_pathlib_path
import click
import json
import requests


@click.group(context_settings=dict(token_normalize_func=str.lower))
def cli():
    """zpy cli

    Zumo Labs cli which is used to create, get, list, upload objects from
    the Zumo Labs backend (ragnarok).
    """
    initialize_config()


@cli.command("help")
def help():
    """display help

    This will display help in order to provide users with more information
    on how to use this tool.
    """
    # TODO: spec this out
    click.echo(
        "zpy - ZumoLabs command line interface\n"
        "\n"
        "zpy is a tool used to list, create, upload, download\n"
        "objects from the ZumoLabs backend (ragnarok)\n"
        "\n"
        "app    - https://app.zumolabs.ai\n"
        "github - https://github.com/ZumoLabs/zpy\n"
        "docs   - https://github.com/ZumoLabs/zpy/tree/main/docs/cli"
    )


@cli.command("env")
@click.argument("env", type=click.Choice(["local", "stage", "prod"]))
def env(env):
    """switch target environment

    This command allows zumo labs developers to swap the endpoint that the
    cli communicates with. Unlikely to be relevant for non-zumo devs.

    Args:
        env (str): new environment for endpoint
    """
    config = read_config()
    old_env, old_endpoint = config["ENVIRONMENT"], config["ENDPOINT"]
    config["ENVIRONMENT"] = env
    config["ENDPOINT"] = get_endpoint(env)
    config["TOKEN"] = None
    write_config(config)
    click.echo("Swapped environment:")
    click.echo(f"  {old_env} -> {config['ENVIRONMENT']}")
    click.echo(f"  {old_endpoint} -> {config['ENDPOINT']}")
    click.echo("zpy login to fetch token")


@cli.command("login")
@click.argument("username", required=True)
@click.password_option(help="The login password.")
def login(username, password):
    """login to ragnarok

    This command will update the zpy config with a token that is fetched
    from the backend using account details.

    Accounts can be created at: app.zumolabs.ai

    Args:
        username (str): developer username
        password (str): developer password
    """
    config = read_config()
    endpoint = f"{config['ENDPOINT']}/auth/login/"
    r = requests.post(endpoint, auth=HTTPBasicAuth(username, password))
    if r.status_code != 200:
        click.secho("Login failed.", err=True, fg="red")
        return
    config["TOKEN"] = r.json()["token"]
    write_config(config)
    click.echo("Login successful!")


@cli.command("config")
def config():
    """display config

    Display current configuration file to developer.
    """
    pretty_config = json.dumps(read_config(), indent=2)
    click.echo(f"Zpy cli configuration:\n{pretty_config}")


@cli.command("version")
def version():
    """version

    Display the zpy cli version.
    """
    import zpy

    click.echo(f"Version: {zpy.__version__}")


# ------- LIST


@cli.group()
def list():
    """list objects

    List group is used for list commands on backend objects.
    """
    pass


@list.command("datasets")
def list_datasets():
    """list datasets

    List datasets from backend.
    """
    from cli.datasets import fetch_datasets

    try:
        with Loader("Fetching datasets..."):
            datasets = fetch_datasets()
        click.echo("Fetched datasets succesfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch datasets {e}.", fg="red", err=True)
        return

    tbl = TableLogger(columns="name,state,type,created,id", default_colwidth=30)
    for d in datasets:
        tbl(d["name"], d["state"].lower(), d["type"], d["created_at"], d["id"])


@list.command("sims")
def list_sims():
    """list datasets

    List sims from backend.
    """
    from cli.sims import fetch_sims

    try:
        with Loader("Fetching sims..."):
            sims = fetch_sims()
        click.echo("Fetched sims succesfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch sims {e}.", fg="red", err=True)
        return

    tbl = TableLogger(
        columns="name,state,zpy_version,blender_version,created", default_colwidth=30
    )
    for s in sims:
        tbl(
            s["name"],
            s["state"],
            s["zpy_version"],
            s["blender_version"],
            s["created_at"],
        )


@list.command("jobs")
def list_jobs():
    """list jobs

    List jobs from backend.
    """
    from cli.jobs import fetch_jobs

    try:
        with Loader("Fetching jobs..."):
            jobs = fetch_jobs()
        click.echo("Fetched jobs succesfully.")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to fetch jobs {e}.", fg="red", err=True)
        return

    tbl = TableLogger(columns="state,name,operation,created", default_colwidth=30)
    for j in jobs:
        tbl(j["state"], j["name"], j["operation"], j["created_at"])


# ------- GET


@cli.group()
def get():
    """get object

    Get group is used for download commands on backend objects.
    """
    pass


@get.command("dataset")
@click.argument("name")
@click.argument("dtype", type=click.Choice(["job", "generated", "uploaded"]))
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
def get_dataset(name, dtype, path):
    """get dataset

    Download dataset from backend.

    Args:
        name (str): name of dataset
        dtype (str): type of dataset
        path (str): directory to put zipped dataset
    """
    from cli.datasets import download_dataset

    try:
        output_path = download_dataset(name, path, dtype)
        click.echo(f"Downloaded {dtype} dataset '{name}' to {output_path}")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to download dataset: {e}", fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to download dataset: {e}", fg="yellow", err=True)


@get.command("sim")
@click.argument("name")
@click.argument(
    "path",
    type=click.Path(exists=True, file_okay=False, writable=True, resolve_path=True),
)
def get_sim(name, path):
    """get sim

    Download sim from backend.

    Args:
        name (str): name of sim
        path (str): directory to put zipped sim
    """
    from cli.sims import download_sim

    try:
        output_path = download_sim(name, path)
        click.echo(f"Downloaded sim '{name}' to {output_path}")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to download sim: {e}", fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to download sim: {e}", fg="yellow", err=True)


# -------  UPLOAD


@cli.group()
def upload():
    """upload object

    Upload group is used for upload commands on backend objects.
    """
    pass


@upload.command("sim")
@click.argument("name")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
def upload_sim(name, path):
    """upload sim

    Upload sim to backend.

    Args:
        name (str): name of sim
        path (str): path to sim
    """
    from cli.sims import create_sim

    if to_pathlib_path(path).suffix != ".zip":
        click.secho(f"File {path} must be of type zip", fg="red", err=True)
    try:
        with Loader("Uploading sim..."):
            create_sim(name, path)
        click.secho(f"Uploaded sim {path} with name '{name}'", fg="green")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to upload sim: {e}", fg="red", err=True)


@upload.command("dataset")
@click.argument("name")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True))
def upload_dataset(name, path):
    """upload dataset

    Upload dataset to backend.

    Args:
        name (str): name of dataset
        path (str): path to dataset
    """
    from cli.datasets import create_uploaded_dataset

    if to_pathlib_path(path).suffix != ".zip":
        click.secho(f"File {path} must be of type zip", fg="red", err=True)
    try:
        with Loader("Uploading dataset..."):
            create_uploaded_dataset(name, path)
        click.secho(f"Uploaded dataset {path} with name '{name}'", fg="green")
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to upload dataset: {e}", fg="red", err=True)


# ------- CREATE


@cli.group()
def create():
    """create object

    Create group is used for create commands on backend objects.
    """
    pass


@create.command("dataset")
@click.argument("name")
@click.argument("sim")
@click.argument("args", nargs=-1)
def create_dataset(name, sim, args):
    """create dataset

    Create a generated dataset object in backend that will trigger
    the generation of the dataset.

    Args:
        name (str): name of new dataset
        sim (str): name of sim dataset is built with
        args (List(str)): configuration of sim for this dataset
    """
    from cli.datasets import create_generated_dataset

    try:
        dataset_config = parse_args(args)
    except Exception:
        click.secho("Failed to parse args: {args}", fg="yellow", err=True)
        return
    try:
        create_generated_dataset(name, sim, parse_args(args))
        click.secho(
            f"Created dataset '{name}' from sim '{sim}' with config {dataset_config}",
            fg="green",
        )
    except requests.exceptions.HTTPError as e:
        click.secho(f"Failed to create dataset: {e}", fg="red", err=True)
    except NameError as e:
        click.secho(f"Failed to create dataset: {e}", fg="yellow", err=True)


@create.command("sweep")
@click.argument("name")
@click.argument("sim")
@click.argument("number")
@click.argument("args", nargs=-1)
def create_sweep(name, sim, number, args):
    """create sweep

    Create a sweep of generated dataset object in backend that will trigger
    the generation of the dataset. Sweep is just a series of create dataset
    calls with different seeds set.

    Args:
        name (str): name of new dataset
        sim (str): name of sim dataset is built with
        number (str): number of datasets to create
        args (List(str)): configuration of sim for this dataset
    """
    from cli.datasets import create_generated_dataset

    try:
        dataset_config = parse_args(args)
    except Exception:
        click.secho("Failed to parse args: {args}", fg="yellow", err=True)
        return
    for i in range(int(number)):
        dataset_name = f"{name} seed{i}"
        dataset_config["seed"] = i
        try:
            create_generated_dataset(dataset_name, sim, dataset_config)
            click.secho(
                f"Created dataset '{dataset_name}' from sim '{sim}' with config {dataset_config}",
                fg="green",
            )
        except requests.exceptions.HTTPError as e:
            click.secho(f"Failed to create dataset: {e}", fg="red", err=True)
        except NameError as e:
            click.secho(f"Failed to create dataset: {e}", fg="yellow", err=True)
            return
    click.echo(f"Finished creating {number} datasets from sim '{sim}'.")


@create.command("job")
@click.argument("name")
@click.argument("operation")
@click.option("filters", "-f", multiple=True)
@click.option(
    "configfile",
    "--configfile",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "sweepfile",
    "--sweepfile",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
def create_job(name, operation, filters, configfile, sweepfile):
    """create job

    Create a job object in backend that will trigger an operation on
    datasets filtered by the filters.

    Args:
        name (str): name of new job
        operation (str): name of operation to run on datasets
        filters (str): string filters for dataset names to run job on
        configfile (str): json configuration for the job
        sweepfile (str): sweep json to launch a suite of jobs
    """
    from cli.datasets import filter_datasets
    from cli.jobs import create_new_job

    datasets = []
    for dfilter in filters:
        try:
            with Loader(f"Filtering datasets by '{dfilter}'..."):
                filtered_datasets = filter_datasets(dfilter)
            filtered_datasets_names = [*filtered_datasets.keys()]
            click.echo(
                f"Filtered datasets by filter '{dfilter}':\n{filtered_datasets_names}"
            )
            datasets.append(filtered_datasets.values())
        except requests.exceptions.HTTPError as e:
            click.secho(f"Failed to filter datsets {e}", fg="red", err=True)

    job_configs = []
    if configfile:
        config = read_json(configfile)
        job_configs.append(config)
        click.echo(f"Parsed config file {configfile} : {config}")
    elif sweepfile:
        sweep_config = read_json(sweepfile)
        try:
            configs = resolve_sweep(sweep_config)
        except Exception as e:
            click.secho(
                f"Failed to resolve sweep file {sweepfile} {e}", fg="yellow", err=True
            )
            return
        job_configs.extend(configs)
        click.echo(f"Parsed sweep file {sweepfile} : {sweep_config}")
    else:
        job_configs.append(dict())

    click.confirm(f"Launch {len(job_configs)} jobs?", abort=True)

    for i, config in enumerate(job_configs):
        job_name = name if i == 0 else f"{name} {i}"
        try:
            create_new_job(job_name, operation, config, datasets)
            click.secho(
                f"Created {operation} job '{job_name}' with config {config}", fg="green"
            )
        except requests.exceptions.HTTPError as e:
            click.secho(f"Failed to create job: {e}", fg="red", err=True)

    click.echo(f"Finished creating {len(job_configs)} jobs with name '{name}'")



from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

import logging as py_logging

log = py_logging.getLogger(__name__)

# Import files first since it is used in other zpy modules
from zpy import files

# Color utilities will reset the random colors on import
from zpy import color

color.reset()
from zpy import gin
from zpy import image
from zpy import logging
from zpy import requests

# Saver object (recording annotations at runtime)
from zpy import saver
from zpy import saver_image
from zpy import saver_video

# Output class (outputting annotations to file)
from zpy import output
from zpy import output_coco
from zpy import output_mot
from zpy import output_zumo
from zpy import output_csv

# If your zpy library has an extra dependency
# which may or may not be installed on a user's
# system, make sure to wrap it in a try-catch
try:
    from zpy import blender
    from zpy import hdris
    from zpy import kdtree
    from zpy import keypoints
    from zpy import camera
    from zpy import material
    from zpy import nodes
    from zpy import objects
    from zpy import render
except ModuleNotFoundError:
    log.debug("Could not load blender utilities.")
try:
    from zpy import viz
except ModuleNotFoundError:
    log.debug("Could not load viz utilities.")



"""
    ZUMO dataset format.
"""
import logging
from pathlib import Path
from typing import Union

import gin
import zpy

log = logging.getLogger(__name__)


class ZUMOParseError(Exception):
    """Invalid ZUMO Annotation found when parsing data contents."""

    pass


@gin.configurable
class OutputZUMO(zpy.output.Output):
    """Output class for ZUMO style annotations.

    The ZUMO format is basically a serialized version of zpy.saver.Saver.

    """

    ANNOTATION_FILENAME = Path("_annotations.zumo.json")

    def __init__(self, *args, **kwargs) -> Path:
        super().__init__(*args, annotation_filename=self.ANNOTATION_FILENAME, **kwargs)

    def output_annotations(
        self,
        annotation_path: Union[Path, str] = None,
    ) -> Path:
        """Output annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.

        Returns:
            Path: Path to annotation file.
        """
        annotation_path = super().output_annotations(annotation_path=annotation_path)
        zumo_dict = {
            "metadata": self.saver.metadata,
            "categories": self.saver.categories,
            "images": self.saver.images,
            "annotations": self.saver.annotations,
        }
        # Write out annotations to file
        zpy.files.write_json(annotation_path, zumo_dict)
        # Verify annotations
        parse_zumo_annotations(
            annotation_file=annotation_path, data_dir=self.saver.output_dir
        )
        return annotation_path


@gin.configurable
def parse_zumo_annotations(
    annotation_file: Union[Path, str],
    data_dir: Union[Path, str] = None,
    output_saver: bool = False,
) -> zpy.saver_image.ImageSaver:
    """Parse COCO annotations, optionally output a ImageSaver object.

    Args:
        annotation_file (Union[Path, str]): Path to annotation file.
        data_dir (Union[Path, str], optional): Directory containing data (images, video, etc).
        output_saver (bool, optional): Whether to return a Saver object or not. Defaults to False.

    Raises:
        ZUMOParseError: Various checks on annotations, categories, images

    Returns:
        zpy.saver.Saver: Saver object.
    """
    log.info(f"Parsing ZUMO annotations at {annotation_file}...")

    # Check annotation file path
    annotation_file = zpy.files.verify_path(annotation_file)
    if data_dir is not None:
        data_dir = zpy.files.verify_path(data_dir, check_dir=True)
    else:
        # If no data_dir, assume annotation file is in the root folder.
        data_dir = annotation_file.parent

    zumo_metadata = zpy.files.read_json(annotation_file)
    images = zumo_metadata["images"]
    if len(images.keys()) == 0:
        raise ZUMOParseError(f"no images found in {annotation_file}")
    categories = zumo_metadata["categories"]
    if len(categories.keys()) == 0:
        raise ZUMOParseError(f"no categories found in {annotation_file}")
    annotations = zumo_metadata["annotations"]
    if len(annotations) == 0:
        raise ZUMOParseError(f"no annotations found in {annotation_file}")
    log.info(
        f"images:{len(images)} categories:{len(categories)} annotations:{len(annotations)}"
    )

    # Optionally output a saver object.
    if output_saver:
        saver = zpy.saver_image.ImageSaver(
            output_dir=data_dir,
            annotation_path=annotation_file,
            description=zumo_metadata["metadata"]["description"],
            clean_dir=False,
        )

    # Check Images
    log.info("Parsing images...")
    img_ids = []
    for image_id, img in images.items():
        # HACK: JSON will convert int keys to str, so undo that here
        image_id = int(image_id)
        # Image ID
        if not image_id == img["id"]:
            raise ZUMOParseError(
                f"image id {image_id} does not match image dict key {img['id']}"
            )
        if not isinstance(image_id, int):
            raise ZUMOParseError(f"image id {image_id} must be int.")
        if image_id in img_ids:
            raise ZUMOParseError(f"image id {image_id} already used.")
        img_ids.append(image_id)
        if image_id < 0:
            raise ZUMOParseError(f"invalid image id {image_id}")
        # Frame
        frame = img.get("frame", None)
        if frame is not None:
            if not isinstance(frame, int):
                raise ZUMOParseError(f"frame {frame} must be int.")
            if image_id < 0:
                raise ZUMOParseError(f"invalid image frame {frame}")
        # Height and Width
        height, width = img["height"], img["width"]
        if not isinstance(height, int):
            raise ZUMOParseError(f"height {height} must be int.")
        if not isinstance(width, int):
            raise ZUMOParseError(f"width {width} must be int.")
        if height <= 0 or width <= 0:
            raise ZUMOParseError(f"width and height h:{height} w:{width} must be > 0")
        # Name
        name = img.get("name", None)
        if name is not None:
            if not isinstance(name, str):
                raise ZUMOParseError(f"name {name} must be str.")
            if (
                frame is not None
                and (not zpy.files.id_from_image_name(name) == frame)
                and (not zpy.files.id_from_image_name(name) == image_id)
            ):
                raise ZUMOParseError(
                    f"name {name} does not correspond to"
                    f" frame {frame} or image_id {image_id}."
                )
        # Output path
        output_path = img.get("output_path", None)
        if output_path is not None:
            if not isinstance(output_path, str):
                raise ZUMOParseError(f"output_path {output_path} must be str.")
            if not Path(output_path).exists():
                raise ZUMOParseError(f"output_path {output_path} does not exist")
        # Save each image to ImageSaver object
        if output_saver:
            saver.images[image_id] = img

    # Check Categories
    log.info("Parsing categories...")
    cat_ids = []
    cat_names = []
    for category_id, category in categories.items():
        # Category Name
        category_name = category["name"]
        if not isinstance(category_name, str):
            raise ZUMOParseError(f"category_name {category_name} must be str.")
        if category_name in cat_names:
            raise ZUMOParseError(f"category_name {category_name} already used")
        cat_names.append(category_name)
        # HACK: JSON will convert int keys to str, so undo that here
        category_id = int(category_id)
        # Category ID
        if not category_id == category["id"]:
            raise ZUMOParseError(
                f"category_id {category_id} does not match category dict key {category['id']}"
            )
        if not isinstance(image_id, int):
            raise ZUMOParseError(f"category_id {category_id} must be int.")
        if category_id in cat_ids:
            raise ZUMOParseError(f"category id {category_id} already used")
        cat_ids.append(category_id)
        # Supercategories
        if category.get("supercategory", None) is not None:
            pass
        if category.get("supercategories", None) is not None:
            pass
        # Subcategories
        if category.get("subcategory", None) is not None:
            pass
        if category.get("subcategories", None) is not None:
            pass
        # Keypoints
        if category.get("keypoints", None) is not None:
            keypoints = category["keypoints"]
            log.info(f"{len(keypoints)} keypoints:{keypoints}")
            if category.get("skeleton", None) is None:
                raise ZUMOParseError(f"skeleton must be present with {keypoints}")
        # Save each category to ImageSaver object
        if output_saver:
            saver.categories[category_id] = category

    # Check Annotations
    log.info("Parsing annotations...")
    ann_ids = []
    for annotation in annotations:
        # IDs
        image_id, category_id, annotation_id = (
            annotation["image_id"],
            annotation["category_id"],
            annotation["id"],
        )
        if image_id not in img_ids:
            raise ZUMOParseError(f"annotation image id {image_id} not in {img_ids}")
        if category_id not in cat_ids:
            raise ZUMOParseError(
                f"annotation category id {category_id} not in {cat_ids}"
            )
        if annotation_id in ann_ids:
            raise ZUMOParseError(f"annotation id {annotation_id} already used")
        ann_ids.append(annotation_id)

        # Bounding Boxes
        bbox = annotation.get("bbox", None)
        if bbox is not None:
            pass

        # Keypoints
        keypoints = annotation.get("num_keypoints", None)
        if keypoints is not None:
            if "keypoints_xyv" in annotation:
                if (
                    len(annotation["keypoints_xyv"])
                    != int(annotation["num_keypoints"]) * 3
                ):
                    raise ZUMOParseError(
                        "keypoints_xyv not correct size {len(keypoints)}"
                    )
            if "keypoints_xyz" in annotation:
                if (
                    len(annotation["keypoints_xyz"])
                    != int(annotation["num_keypoints"]) * 3
                ):
                    raise ZUMOParseError(
                        "keypoints_xyz not correct size {len(keypoints)}"
                    )

        # Save each annotation to ImageSaver object
        if output_saver:
            saver.annotations.append(annotation)

    if output_saver:
        return saver
    else:
        return None



"""
    Node utilities.
"""
import logging

import bpy
from typing import Tuple

import gin

log = logging.getLogger(__name__)


@gin.configurable
def get_or_make(
    name: str,
    node_type: str,
    tree: bpy.types.NodeTree,
    label_tag: str = "(zpy) ",
    pos: Tuple[float] = None,
) -> bpy.types.Node:
    """Verify existence or create a node.

    Args:
        name (str): Name of the node.
        node_type (str): Node type e.g. "ShaderNodeBackground"
        tree (bpy.types.NodeTree): Node tree where this node will be added.
        label_tag (str, optional): Node label will include this tag to make it easier to identify within Blender.
            Defaults to '(zpy) '.
        pos (Tuple[float], optional): Location of the node in node editor. Defaults to None.

    Returns:
        bpy.types.Node: The newly created (or already existing) node.
    """
    node = tree.nodes.get(name, None)
    if node is None:
        node = tree.nodes.new(node_type)
        node.name = name
    node.label = f"{label_tag}{name}"
    node.bl_description = "This node has been created and/or modified by zpy"
    if pos is not None:
        node.location = pos
    return node


def toggle_nodegroup(node_tree: bpy.types.NodeTree, state: bool = False) -> None:
    """Change the state of all the nodes inside a node group

    Args:
        node_tree (bpy.types.NodeTree): Top level node group
        state: (bool): Toggle state
    """
    for n in node_tree.nodes:
        n.mute = state



"""
    MOT (Multi-Object Tracking) dataset format.

    https://motchallenge.net/faq/

"""
import logging
from pathlib import Path
from typing import Union

import gin
import zpy

log = logging.getLogger(__name__)


class MOTParseError(Exception):
    """Invalid MOT Annotation found when parsing data contents."""

    pass


@gin.configurable
class OutputMOT(zpy.output.Output):
    """Output class for MOT (Multi-Object Tracking) style annotations.

    https://motchallenge.net/faq/

    """

    ANNOTATION_FILENAME = Path("_annotations.mot.csv")

    def __init__(self, *args, **kwargs) -> Path:
        super().__init__(*args, annotation_filename=self.ANNOTATION_FILENAME, **kwargs)

    @gin.configurable
    def output_annotations(
        self,
        annotation_path: Union[Path, str] = None,
    ) -> Path:
        """Output MOT (Multi-Object Tracking) annotations to file.

        Args:
            annotation_path (Union[Path, str], optional): Output path for annotation file.

        Returns:
            Path: Path to annotation file.
        """
        annotation_path = super().output_annotations(annotation_path=annotation_path)
        mot = []
        for annotation in self.saver.annotations:
            if self.saver.images[annotation["image_id"]]["style"] != "default":
                # MOT annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            person_id = annotation.get("person_id", None)
            bbox = annotation.get("bbox", None)
            if (person_id is not None) and (bbox is not None):
                # Each CSV row will have 9 entries
                row = [0] * 9
                # Frame at which the object is present
                row[0] = annotation["frame_id"]
                # Pedestrian trajectory is identiﬁed by a unique ID
                row[1] = person_id
                # Coordinate of the top-left corner of the pedestrian bounding box
                row[2] = bbox[0]
                # Coordinate of the top-left corner of the pedestrian bounding box
                row[3] = bbox[1]
                # Width in pixels of the pedestrian bounding box
                row[4] = bbox[2]
                # Height in pixels of the pedestrian bounding box
                row[5] = bbox[3]
                # Flag whether the entry is to be considered (1) or ignored (0).
                row[6] = 1
                # TODO: Type of object annotated
                """
                MOT Types:
                    Pedestrian 1
                    Person on vehicle 2
                    Car 3
                    Bicycle 4
                    Motorbike 5
                    Non motorized vehicle 6
                    Static person 7
                    Distractor 8
                    Occluder 9
                    Occluder on the ground 10
                    Occluder full 11
                    Reflection 12
                    Crowd 13
                """
                row[7] = annotation["mot_type"]
                # TODO: Visibility ratio, a number between 0 and 1 that says how much of that object
                # is visible. Can be due to occlusion and due to image border cropping.
                row[8] = 1.0
                # Add to mot list
                mot.append(row)
        # Write out annotations to file
        zpy.files.write_csv(annotation_path, mot)
        # Verify annotations
        parse_mot_annotations(annotation_path)
        return annotation_path


@gin.configurable
def parse_mot_annotations(
    annotation_file: Union[Path, str],
) -> None:
    """Parse MOT annotations.

    Args:
        annotation_file (Union[Path, str]): Path to annotation file.

    Raises:
        MOTParseError: Rows are not length 9.
    """
    log.info(f"Verifying MOT annotations at {annotation_file}...")
    mot = zpy.files.read_csv(annotation_file)
    for row in mot:
        if not len(row) == 9:
            raise MOTParseError(
                f"Each row in MOT csv must have len 9, found len {len(row)}"
            )
    # TODO: Return Saver object.



"""
    Video version of Saver object.
"""
import logging
from pathlib import Path
from typing import Dict, Union
from datetime import timedelta

import gin

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class VideoSaver(zpy.saver.Saver):
    """Holds the logic for saving video annotations at runtime."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.videos = {}
        self.video_name_to_id = {}

    @gin.configurable
    def add_video(
        self,
        name: str = "default video",
        style: str = "default",
        output_path: Union[Path, str] = "/tmp/test.avi",
        width: int = 640,
        height: int = 480,
        length: timedelta = 0,
        zero_indexed: bool = True,
        **kwargs,
    ) -> Dict:
        """Add a new annotation to the Saver object.

        Args:
            name (str, optional): Unique video name. Defaults to 'default video'.
            style (str, optional): Type of image in [default, segmenation, depth]. Defaults to 'default'.
            output_path (Union[Path, str], optional): Path to video file. Defaults to '/tmp/test.avi'.
            width (int, optional): Width of video frame. Defaults to 640.
            height (int, optional): Height of video frame. Defaults to 480.
            length (timedelta, optional): Length of video in seconds. Defaults to 0.
            zero_indexed (bool, optional): Whether video id is zero indexed. Defaults to True.

        Returns:
            Dict: The video annotation dictionary.
        """
        video = {
            "name": name,
            "style": style,
            "output_path": str(output_path),
            "relative_path": str(Path(output_path).relative_to(self.output_dir)),
            "width": width,
            "height": height,
            "length": length,
        }
        video.update(**kwargs)
        video["id"] = len(self.videos.keys())
        video["id"] += 0 if zero_indexed else 1
        log.debug(f"Adding video: {zpy.files.pretty_print(video)}")
        self.videos[video["id"]] = video
        self.video_name_to_id[name] = video["id"]
        return video

    @gin.configurable
    def add_annotation(
        self,
        *args,
        video: str = "default video",
        **kwargs,
    ) -> Dict:
        """Add a new annotation to the Saver object.

        Args:
            video (str, optional): Unique video name. Defaults to 'default video'.

        Returns:
            Dict: The annotation dictionary.
        """
        annotation = super().add_annotation(*args, **kwargs)
        video_id = self.video_name_to_id.get(video, None)
        assert video_id is not None, f"Could not find id for video {video}"
        annotation["video_id"] = video_id
        annotation.update(**kwargs)
        self.annotations.append(annotation)
        return annotation

    @gin.configurable
    def output_meta_analysis(self):
        """Perform a full meta analysis, outputting some meta files."""
        # TODO: implement meta-analysis for video datasets
        pass



from cli.utils import fetch_auth
import requests
import json


@fetch_auth
def create_new_job(name, operation, config, datasets, url, auth_headers):
    """create job

    Create a job object on ZumoLabs backend. This will trigger the backend
    to run the job.

    Args:
        name (str): name of job
        operation (str): job type
        config (dict): configuration for job
        datasets (dict): list of dataset ids
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/jobs/"
    data = {
        "operation": operation.upper(),
        "name": name,
        "input_data_sets": datasets,
        "config": json.dumps(config),
    }
    r = requests.post(endpoint, data=data, headers=auth_headers)
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def fetch_jobs(url, auth_headers):
    """fetch jobs

    Fetch job objects from ZumoLabs backend.

    Args:
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        list: list of jobs
    """
    endpoint = f"{url}/api/v1/jobs/"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]



"""
    Output panel and functions.
"""
import importlib
import logging

import bpy
import zpy
from bpy.types import Operator

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """Properties applied to scenes."""
    bpy.types.Scene.zpy_output_path = bpy.props.StringProperty(
        name="Output Path",
        description="Output path for rendered images, annotations, etc.",
        default=str(zpy.files.default_temp_path()),
        subtype="DIR_PATH",
    )


class CleanOutputDirOperator(bpy.types.Operator):
    """Clean up output dir."""

    bl_idname = "scene.zpy_cleanup_output_dir"
    bl_label = "Clean Output Dir"
    bl_description = "Clean output dir."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        zpy.files.clean_dir(context.scene.zpy_output_path, keep_dir=True)
        return {"FINISHED"}


class OpenOutputDirOperator(Operator):
    """Open file browser at output dir."""

    bl_idname = "scene.zpy_open_output_dir"
    bl_label = "Open Output Dir"
    bl_description = "Open file browser at output dir."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        zpy.files.open_folder_in_explorer(context.scene.zpy_output_path, make=True)
        return {"FINISHED"}


class SCENE_PT_OutputPanel(bpy.types.Panel):
    """UI for the addon that is visible in Blender."""

    bl_idname = "SCENE_PT_OutputPanel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Output Path"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.prop(scene, "zpy_output_path", text="")
        row = layout.row()
        row.operator(
            "scene.zpy_open_output_dir",
            text="Open",
            icon="FILEBROWSER",
        )
        row.operator(
            "scene.zpy_cleanup_output_dir",
            text="Clean",
            icon="PACKAGE",
        )



from cli.config import read_config
from copy import deepcopy
from itertools import product
from tqdm import tqdm
from urllib.request import urlopen
import functools


def parse_filter(str_filter):
    """parse filter

    Parse filter string to field, pattern, regex.
    Valid patterns depend on field but can include exact, startswith, contains,
    iexact, istartswith, icontains. If no pattern or field is provided the defaults
    are name for field and startswith for pattern.

    Examples:
        icontains:hello
        name:icontains:foobar
        state:iexact:ready
        foo

    Args:
        filter (str): filter query

    Return:
        field: field to filter on
        pattern: pattern to apply regex
        regex: string regex for pattern
    """
    filter_arr = str_filter.split(":")
    field, pattern, regex = "name", "startswith", filter_arr[-1]
    if len(filter_arr) == 2:
        pattern = filter_arr[0]
    elif len(filter_arr) == 3:
        field, pattern = filter_arr[0], filter_arr[1]
    return field, pattern, regex


def resolve_sweep(sweep_config):
    """resolve sweep

    Resolve a dictionary into a sweep of dictionaries. Gin bindings
    are used to communicate with job code and therefore the sweep is done
    across gin_bindings in the sweep_config.

    Example:
        {'foo': ['a', 'b']} -> {'foo': 'a'} {'foo': 'b'}

    Args:
        sweep_config (dict): dictionary to unfold into sweep

    Returns:
        list: list of configs resolved from sweep config
    """
    configs, bindings = [], sweep_config["gin_bindings"]
    for random_binding in [dict(zip(bindings, v)) for v in product(*bindings.values())]:
        config = deepcopy(sweep_config)
        config["gin_bindings"] = random_binding
        configs.append(config)
    return configs


def parse_args(args):
    """parse args

    Used by cli to parse arguments passed to cli calls. Includes
    safe eval function to convert from string to other types.

    Example:
        foo 1 bar 2 -> {'foo': 1, 'bar': 2}
    """

    def _safe_eval(key):
        try:
            return eval(key)
        except Exception:
            return key

    keys = args[::2]
    vals = map(lambda x: _safe_eval(x), args[1::2])
    return dict(zip(keys, vals))


def download_url(url, output_path):
    """download url

    Download from url to give output path and visualize using tqdm.

    Args:
        url (str): url to download
        output_path (str): path to download file to
    """
    u = urlopen(url)
    h = u.info()
    totalSize = int(h["Content-Length"])

    fp = open(output_path, "wb")

    blockSize = 8192
    with tqdm(total=totalSize) as pbar:
        while True:
            chunk = u.read(blockSize)
            if not chunk:
                break
            fp.write(chunk)
            pbar.update(blockSize)

    fp.flush()
    fp.close()


def fetch_auth(func):
    """fetch authetication

    Decorator to wrap functions providing the backend url and the
    correct authorization headers for requests.

    Args:
        func: function to wrap

    Returns:
        wrapped function
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = read_config()
        endpoint = config["ENDPOINT"]
        auth_header = {"Authorization": "token {}".format(config["TOKEN"])}
        return func(*args, **kwargs, url=endpoint, auth_headers=auth_header)

    return wrapper



"""
    Utilities for Rendering in Blender.
"""
import logging
import os
import time
from pathlib import Path
from typing import Union, Tuple

import bpy
import gin

import zpy

log = logging.getLogger(__name__)


@gin.configurable
def make_aov_pass(
    style: str = "instance",
) -> None:
    """Make AOV pass in Cycles."""
    scene = zpy.blender.verify_blender_scene()
    # Make sure engine is set to Cycles
    if not (scene.render.engine == "CYCLES"):
        log.warning(" Setting render engine to CYCLES to use AOV")
        scene.render.engine = "CYCLES"
        scene.render.use_compositing = True
    # Only certain styles are available
    valid_styles = ["instance", "category"]
    assert (
        style in valid_styles
    ), f"Invalid style {style} for AOV Output Node, must be in {valid_styles}."
    # Go through existing passes and make sure it doesn't exist before creating
    view_layer = zpy.blender.verify_view_layer()
    for aov in view_layer.aovs.values():
        if aov.name == style:
            log.info(f"AOV pass for {style} already exists.")
            return
    bpy.ops.scene.view_layer_add_aov()
    view_layer.aovs[-1].name = style
    view_layer.update()
    log.info(f"Created AOV pass for {style}.")


@gin.configurable
def make_aov_file_output_node(
    style: str = "rgb",
    add_hsv: bool = True,
    add_lens_dirt: bool = False,
) -> bpy.types.CompositorNodeOutputFile:
    """Make AOV Output nodes in Composition Graph."""
    log.info(f"Making AOV output node for {style}")

    # Only certain styles are available
    valid_styles = ["rgb", "depth", "instance", "category"]
    assert (
        style in valid_styles
    ), f"Invalid style {style} for AOV Output Node, must be in {valid_styles}."

    # Make sure scene composition is using nodes
    scene = zpy.blender.verify_blender_scene()
    scene.use_nodes = True
    tree = scene.node_tree

    # Remove Composite Node if it exists
    composite_node = tree.nodes.get("Composite")
    if composite_node is not None:
        tree.nodes.remove(composite_node)

    rl_node = zpy.nodes.get_or_make("Render Layers", "CompositorNodeRLayers", tree)

    # Instance and category require an AOV pass
    if style in ["instance", "category"]:
        zpy.render.make_aov_pass(style)

    # Visualize node shows image in workspace
    view_node = zpy.nodes.get_or_make(f"{style} Viewer", "CompositorNodeViewer", tree)

    # File output node renders out image
    fileout_node = zpy.nodes.get_or_make(
        f"{style} Output", "CompositorNodeOutputFile", tree
    )
    fileout_node.mute = False

    # HACK: Depth requires normalization node between layer and output
    if style == "depth":

        # Normalization node
        norm_node = zpy.nodes.get_or_make(
            f"{style} Normalize", "CompositorNodeNormalize", tree
        )

        # Negative inversion
        invert_node = zpy.nodes.get_or_make(
            f"{style} Negate", "CompositorNodeInvert", tree
        )

        # Link up the nodes
        tree.links.new(rl_node.outputs["Depth"], norm_node.inputs[0])
        tree.links.new(norm_node.outputs[0], invert_node.inputs["Color"])
        tree.links.new(invert_node.outputs[0], view_node.inputs["Image"])
        tree.links.new(invert_node.outputs[0], fileout_node.inputs["Image"])
    elif style == "rgb":
        _node = rl_node
        if add_lens_dirt:
            _node = lens_dirt_node(node_tree=tree, input_node=rl_node)
            tree.links.new(rl_node.outputs["Image"], _node.inputs["Image"])
        if add_hsv:
            _node = hsv_node(node_tree=tree, input_node=rl_node)
            tree.links.new(rl_node.outputs["Image"], _node.inputs["Image"])
        tree.links.new(_node.outputs["Image"], view_node.inputs["Image"])
        tree.links.new(_node.outputs["Image"], fileout_node.inputs["Image"])
    else:  # category and instance segmentation
        tree.links.new(rl_node.outputs[style], view_node.inputs["Image"])
        tree.links.new(rl_node.outputs[style], fileout_node.inputs["Image"])

    return fileout_node


def hsv_node(
    node_tree: bpy.types.NodeTree,
    input_node: bpy.types.Node,
) -> bpy.types.Node:
    """Adds a Hue-Saturation-Value Node."""
    hsv_node = zpy.nodes.get_or_make("HSV", "CompositorNodeHueSat", node_tree)
    node_tree.links.new(input_node.outputs["Image"], hsv_node.inputs["Image"])
    return hsv_node


def lens_dirt_node(
    node_tree: bpy.types.NodeTree,
    input_node: bpy.types.Node,
) -> bpy.types.Node:
    """TODO: Add lens dirt effect to a compositor node."""
    log.warn("NotImplemented: lens dirt ")
    return input_node


@gin.configurable
def render(
    rgb_path: Union[Path, str] = None,
    depth_path: Union[Path, str] = None,
    iseg_path: Union[Path, str] = None,
    cseg_path: Union[Path, str] = None,
    width: int = 640,
    height: int = 480,
    hsv: Tuple[float] = None,
):
    """Render images using AOV nodes."""
    scene = zpy.blender.verify_blender_scene()
    scene.render.resolution_x = width
    scene.render.resolution_y = height
    scene.cycles.resolution_x = width
    scene.cycles.resolution_y = height
    scene.render.resolution_percentage = 100
    scene.render.image_settings.file_format = "PNG"

    # HACK: Prevents adding frame number to filename
    scene.frame_end = scene.frame_current
    scene.frame_start = scene.frame_current
    scene.render.use_file_extension = False
    scene.render.use_stamp_frame = False
    scene.render.filepath = ""

    # Create AOV output nodes
    render_outputs = {
        "rgb": rgb_path,
        "depth": depth_path,
        "instance": iseg_path,
        "category": cseg_path,
    }
    for style, output_path in render_outputs.items():
        if output_path is not None:
            # Create output node if it is not in scene
            if not scene.use_nodes:
                scene.use_nodes = True
            output_node = scene.node_tree.nodes.get(f"{style} Output", None)
            if output_node is None:
                output_node = make_aov_file_output_node(style=style)
            output_node.base_path = str(output_path.parent)
            output_node.file_slots[0].path = str(output_path.name)
            output_node.format.file_format = "PNG"
            output_node.format.color_mode = "RGB"
            if style in ["rgb"]:
                output_node.format.color_depth = "8"
                output_node.format.view_settings.view_transform = "Filmic"
                if hsv is not None:
                    hsv_node = scene.node_tree.nodes.get("HSV", None)
                    if hsv_node is not None:
                        hsv_node.inputs[1].default_value = max(0, min(hsv[0], 1))
                        hsv_node.inputs[2].default_value = max(0, min(hsv[1], 2))
                        hsv_node.inputs[3].default_value = max(0, min(hsv[2], 2))
                    else:
                        log.warn("Render given HSV but no HSV node found.")
            if style in ["depth"]:
                output_node.format.color_depth = "8"
                output_node.format.use_zbuffer = True
            if style in ["instance", "category"]:
                output_node.format.color_depth = "8"
                output_node.format.view_settings.view_transform = "Raw"
            log.debug(f"Output node for {style} image pointing to {str(output_path)}")

    if render_outputs.get("rgb", None) is not None:
        # Mute segmentation and depth output nodes
        _mute_aov_file_output_node("category", mute=True)
        _mute_aov_file_output_node("instance", mute=True)
        _mute_aov_file_output_node("depth", mute=True)
        _mute_aov_file_output_node("rgb", mute=False)
        default_render_settings()
        _render()

    cseg_is_on = render_outputs.get("category", None) is not None
    iseg_is_on = render_outputs.get("instance", None) is not None
    depth_is_on = render_outputs.get("depth", None) is not None
    if cseg_is_on or iseg_is_on or depth_is_on:
        # Un-mute segmentation and depth output nodes
        _mute_aov_file_output_node("category", mute=(not cseg_is_on))
        _mute_aov_file_output_node("instance", mute=(not iseg_is_on))
        _mute_aov_file_output_node("depth", mute=(not depth_is_on))
        _mute_aov_file_output_node("rgb", mute=True)
        segmentation_render_settings()
        _render()

    # Save intermediate blenderfile
    if log.getEffectiveLevel() == logging.DEBUG:
        # HACK: Use whatever output path is not None
        for style, output_path in render_outputs.items():
            if output_path is not None:
                break
        _filename = f"_debug.post.{output_path.stem}.blend"
        _path = output_path.parent / _filename
        zpy.blender.save_debug_blenderfile(_path)

    # HACK: Rename image outputs due to stupid Blender reasons
    for style, output_path in render_outputs.items():
        if output_path is not None:
            _bad_name = str(output_path) + "%04d" % scene.frame_current
            os.rename(_bad_name, str(output_path))
            log.info(f"Rendered {style} image saved to {str(output_path)}")


# TODO: Eventually remove this deprecated function name
def render_aov(*args, **kwargs):
    return render(*args, **kwargs)


def _mute_aov_file_output_node(style: str, mute: bool = True):
    """Mute (or un-mute) an AOV output node for a style."""
    log.debug(f"Muting AOV node for {style}")
    scene = zpy.blender.verify_blender_scene()
    node = scene.node_tree.nodes.get(f"{style} Output", None)
    if node is not None:
        node.mute = mute


@gin.configurable
def default_render_settings(
    samples: int = 96,
    tile_size: int = 48,
    spatial_splits: bool = False,
    is_aggressive: bool = False,
) -> None:
    """Render settings for normal color images.

    Args:
        samples (int, optional): Number of Cycles samples per frame
        tile_size (int, optional): Rendering tile size in pixel dimensions
        spatial_splits (bool, optional): Toogle for BVH split acceleration
        is_aggressive (bool, optional): Toogles aggressive render time reduction settings
    """
    scene = zpy.blender.verify_blender_scene()
    # Make sure engine is set to Cycles
    if not (scene.render.engine == "CYCLES"):
        log.warning(" Setting render engine to CYCLES")
        scene.render.engine == "CYCLES"

    scene.cycles.samples = samples
    scene.cycles.use_adaptive_sampling = True
    scene.cycles.use_denoising = False
    scene.cycles.denoiser = "OPENIMAGEDENOISE"

    scene.render.film_transparent = False
    scene.render.dither_intensity = 1.0
    scene.render.filter_size = 1.5

    view_layer = zpy.blender.verify_view_layer()
    scene.render.use_single_layer = True
    view_layer.pass_alpha_threshold = 0.5

    scene.cycles.max_bounces = 12
    scene.cycles.diffuse_bounces = 4
    scene.cycles.glossy_bounces = 4
    scene.cycles.transparent_max_bounces = 4
    scene.cycles.transmission_bounces = 12

    scene.cycles.sample_clamp_indirect = 2.5
    scene.cycles.sample_clamp_direct = 2.5
    scene.cycles.blur_glossy = 1
    scene.cycles.caustics_reflective = False
    scene.cycles.caustics_refractive = False

    scene.view_settings.view_transform = "Filmic"
    scene.display.render_aa = "8"
    scene.display.viewport_aa = "FXAA"
    scene.display.shading.color_type = "TEXTURE"
    scene.display.shading.light = "STUDIO"
    scene.display.shading.show_specular_highlight = True

    scene.render.tile_x = tile_size
    scene.render.tile_y = tile_size
    scene.cycles.debug_use_spatial_splits = spatial_splits
    scene.render.use_persistent_data = True

    if is_aggressive:
        scene.cycles.samples = 64

        scene.cycles.max_bounces = 8
        scene.cycles.diffuse_bounces = 2
        scene.cycles.glossy_bounces = 2
        scene.cycles.transparent_max_bounces = 2
        scene.cycles.transmission_bounces = 4

        scene.render.use_simplify = True
        scene.cycles.ao_bounces_render = 1
        scene.world.light_settings.use_ambient_occlusion = True
        scene.world.light_settings.distance = 40
        scene.world.light_settings.ao_factor = 0.5


def segmentation_render_settings():
    """Render settings for segmentation images."""
    scene = zpy.blender.verify_blender_scene()

    # Make sure engine is set to Cycles
    if not (scene.render.engine == "CYCLES"):
        log.warning(" Setting render engine to CYCLES")
        scene.render.engine == "CYCLES"

    scene.render.film_transparent = True
    scene.render.dither_intensity = 0.0
    scene.render.filter_size = 0.0

    scene.cycles.samples = 1
    scene.cycles.diffuse_bounces = 0
    scene.cycles.diffuse_samples = 0

    view_layer = zpy.blender.verify_view_layer()
    view_layer.pass_alpha_threshold = 0.0

    scene.cycles.max_bounces = 0
    scene.cycles.bake_type = "EMIT"
    scene.cycles.use_adaptive_sampling = False
    scene.cycles.use_denoising = False
    scene.cycles.denoising_radius = 0

    scene.view_settings.view_transform = "Raw"

    scene.display.render_aa = "OFF"
    scene.display.viewport_aa = "OFF"
    scene.display.shading.color_type = "MATERIAL"
    scene.display.shading.light = "FLAT"
    scene.display.shading.show_specular_highlight = False


def _render(
    threads: int = 4,
    logfile_path: Union[Path, str] = "blender_render.log",
) -> None:
    """The actual call to render a frame in Blender.

    Args:
        threads (int, optional): Number of threads to render on. Defaults to 4.
        logfile_path (Union[Path, str]): Path to save render logfile.
    """
    start_time = time.time()
    scene = zpy.blender.verify_blender_scene()
    # TODO: Get a better default number based on number of available cores
    scene.render.threads = threads
    # TODO: The commented out code here only works on Linux (fails on Windows)
    # try:
    #     # HACK: This disables the blender log by redirecting output to log file
    #     # https://blender.stackexchange.com/questions/44560
    #     open(logfile, 'a').close()
    #     old = os.dup(1)
    #     sys.stdout.flush()
    #     os.close(1)
    #     os.open(logfile, os.O_WRONLY)
    # except Exception as e:
    #     log.warning(f'Render log removal raised exception {e}')
    try:
        # This is the actual render call
        bpy.ops.render.render()
    except Exception as e:
        log.warning(f"Render raised exception {e}")
    # try:
    #     # disable output redirection
    #     os.close(1)
    #     os.dup(old)
    #     os.close(old)
    # except Exception as e:
    #     log.warning(f'Render log removal raised exception {e}')
    duration = time.time() - start_time
    log.info(f"Rendering took {duration}s to complete.")



"""
    Gin-config utilities.
"""
import logging
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, Tuple, Union

import gin
import zpy

log = logging.getLogger(__name__)


def replace_human_redable_kwargs(
    gin_bindings: Dict,
    human_conversion: Dict = {
        "seed": "zpy.blender.set_seed.seed",
        "output_dir": "zpy.saver.Saver.output_dir",
        "output_path": "zpy.saver.Saver.output_dir",
        "num_images": "zpy.blender.step.num_steps",
        "num_frames": "zpy.blender.step.num_steps",
    },
) -> Tuple[str, Any]:
    """[summary]

    Args:
        gin_bindings (Dict): Gin bindings dictionary {gin binding : value}.
        human_conversion (Dict, optional): Conversion from zpy keys to human-readable keys.

    Returns:
        Tuple[str, Any]: A single gin bindings.

    Yields:
        Iterator[Tuple[str, Any]]: New gin bindings.
    """
    log.info("Converting human readable bindings to gin...")
    for key, value in gin_bindings.items():
        if human_conversion.get(key, None) is not None:
            log.info(f"Converted {key} to {human_conversion[key]}")
            yield human_conversion[key], value
        else:
            yield key, value


def parse_gin_bindings(
    gin_bindings: Dict = None,
) -> None:
    """Parse any extra gin bindings to the config.

    Args:
        gin_bindings (Dict, optional): Gin bindings dictionary {gin binding : value}.
    """
    if gin_bindings is None:
        log.info("No additional gin bindings to parse")
    else:
        log.info(f"Parsing additional bindings: {pformat(gin_bindings)}")
        with gin.unlock_config():
            for key, value in replace_human_redable_kwargs(gin_bindings):
                try:
                    gin.bind_parameter(key, value)
                    _message = "BOUND  "
                except Exception:
                    _message = "IGNORED"
                log.info(f"{_message} - {key} : {value}")


def parse_gin_config(
    gin_config: str = None,
    gin_config_dir: Union[Path, str] = "$CONFIG",
) -> None:
    """Parse a gin config file by path.

    Args:
        gin_config (str, optional): Name of gin config.
        gin_config_dir (Union[Path, str], optional): Directory with gin configs.

    Raises:
        zpy.requests.InvalidRequest: Cannot find gin config at path.
    """
    if gin_config is None:
        log.info("No gin file to parse.")
    else:
        if not gin_config.endswith(".gin"):
            gin_config = gin_config + ".gin"
        gin_config_filename = Path(gin_config)
        gin_config_dir = zpy.files.verify_path(gin_config_dir, check_dir=True)
        gin_config_path = gin_config_dir / gin_config_filename
        log.info(f"Parsing gin config at {gin_config_path}")
        if not gin_config_path.exists():
            raise zpy.requests.InvalidRequest(
                f"Could not find gin config at {gin_config_path}"
            )
        gin.clear_config()
        gin.parse_config_file(str(gin_config_path))


def parse_gin_in_request(
    request: Dict,
) -> None:
    """Parse any gin related keys in a request dict.

    Args:
        request (Dict): Request dictionary (see zpy.requests).
    """
    zpy.gin.parse_gin_config(gin_config=request.get("gin_config", None))
    zpy.gin.parse_gin_bindings(gin_bindings=request.get("gin_bindings", None))
    gin.finalize()



"""
    Machine Learning Experiment Hooks.
"""
from typing import Dict, Union
from pathlib import Path
import json
import requests
import logging

ENDPOINT = "https://ragnarok.zumok8s.org/api/v1/experiment/"
experiment = None
logger = None


def init(
    name: str,
    sim: str = None,
    dataset: str = None,
    config: Dict = None,
    api_key: str = None,
) -> None:
    """Initialize a experiment run.

    Args:
        name (str): identifier for experiment
        sim (str. optional): identifier for simulation associated with experiment
        dataset (str, optional): identifier for dataset associated with experiment
        config (dict, optional): configuration details about experiment
        api_key (str, required): api_key to auth with backend
    Returns:
        experiment object
    """
    if api_key is None:
        raise PermissionError("please input zpy api_key")
    global logger
    logger = logging.getLogger(__name__)
    exp = Experiment(
        name=name, sim=sim, dataset=dataset, config=config, api_key=api_key
    )
    global experiment
    experiment = exp
    exp._create()


def log(
    metrics: str = None,
    file_path: str = None,
) -> None:
    """Log an update to experiment.

    Args:
        metrics (str, optional): free form dictionary of data to log
        file_path (str, optional): file path to upload

    Raises:
        FileNotFoundError: if file_path doesnt exist
    """
    global experiment
    exp = experiment
    if file_path:
        file_path = Path(file_path).resolve()
    exp._update(file_path=file_path, metrics=metrics)


class Experiment:
    """experiment class for uploading"""

    def __init__(
        self,
        name: str = None,
        sim: str = None,
        dataset: str = None,
        config: Dict = None,
        api_key: str = None,
    ) -> None:
        """Experiment Class for ragnarok upload.

        Args:
            name (str): identifier for experiment
            sim (str. optional): identifier for simulation associated with experiment
            dataset (str, optional): identifier for dataset associated with experiment
            config (dict, optional): configuration details about experiment
            api_key (str, required): api_key to auth with backend
        """
        self.name = name
        self.sim = sim
        self.dataset = dataset
        self.config = config
        self.auth_headers = {"Authorization": "token {}".format(api_key)}
        self.id = None

    def _post(self, data=None):
        """post to endpoint"""
        r = requests.post(ENDPOINT, data=data, headers=self.auth_headers)
        if r.status_code != 201:
            logger.debug(f"{r.text}")
            r.raise_for_status()
        self.id = json.loads(r.text)["id"]
        logger.debug(f"{r.status_code}: {r.text}")

    def _put(self, data=None, files=None):
        """put to endpoint"""
        r = requests.put(
            f"{ENDPOINT}{self.id}/", data=data, files=files, headers=self.auth_headers
        )
        if r.status_code != 200:
            logger.debug(f"{r.text}")
            r.raise_for_status()
        logger.debug(f"{r.status_code}: {r.text}")

    def _create(self):
        """request to create experiment"""
        data = {"name": self.name}
        if self.sim:
            data["sim_name"] = self.sim
        if self.dataset:
            data["data_set_name"] = self.dataset
        if self.config:
            data["config"] = json.dumps(self.config)
        self._post(data=data)

    def _update(self, file_path: Union[Path, str] = None, metrics: Dict = None) -> None:
        """request to update experiment"""
        data = {"name": self.name}
        if metrics:
            data["metrics"] = json.dumps(metrics)
        files = None
        if file_path:
            files = {"file": open(file_path, "rb")}
            data["file_name"] = Path(file_path).name
        self._put(data=data, files=files)



"""
    Blender utilities.
"""
import inspect
import logging
import os
import random
import time
from functools import wraps
from pathlib import Path
from typing import Dict, List, Union

import bpy
import mathutils
import numpy as np

import gin
import zpy

log = logging.getLogger(__name__)


def use_gpu(
    compute_device_type="CUDA",
    use_cpu=True,
) -> None:
    """Choose the rendering devices for rendering.

    The hybrid render device options (GPU+CPU) are possible for CUDA and OPTIX

    Args:
        compute_device_type (str, optional): One of [NONE, CUDA, OPTIX, OPENCL]. Defaults to 'CUDA'.
        use_cpu (bool, optional): Use CPU with compute device. Defaults to True.

    Raises:
        RuntimeError: Compute device is not a valid choice.
    """
    C = bpy.context
    preferences = bpy.context.preferences
    cycles_preferences = preferences.addons["cycles"].preferences
    compute_devices = [d[0] for d in cycles_preferences.get_device_types(C)]
    if compute_device_type not in compute_devices:
        raise RuntimeError("Non-existing device type")
    else:
        cycles_preferences.compute_device_type = compute_device_type
        devices = cycles_preferences.get_devices_for_type(compute_device_type)
        if len(devices) > 0:
            for c in devices:
                c.use = True
                if c.type == "CPU":
                    c.use = use_cpu
                log.info(f"Using devices {c} {c.type} {c.use}")

    C.scene.cycles.device = "GPU"
    log.info(f"Using gpu type:{compute_device_type} cpu:{use_cpu}")


@gin.configurable
def set_seed(
    seed: int = 0,
) -> None:
    """Set the random seed (sets the python and numpy seed).

    Args:
        seed (int, optional): Integer seed. Defaults to 0.
    """
    if log.getEffectiveLevel() == logging.DEBUG:
        # When debugging you want to run into errors related
        # to specific permutations of the random variables, so
        # you need to vary the seed to run into them.
        seed = random.randint(1, 100)
    log.info(f"Setting random seed to {seed}")
    random.seed(seed)
    np.random.seed(seed)
    mathutils.noise.seed_set(seed)


@gin.configurable
def step(
    num_steps: int = 3,
    framerate: int = 0,
    start_frame: int = 1,
    refresh_ui: bool = False,
) -> int:
    """Steps the sim forward (Blender frames).

    Args:
        num_steps (int, optional): Number of steps to take before the yield stops. Defaults to 16.
        framerate (int, optional): How many Blender frames to move forward in each step. Defaults to 0.
        start_frame (int, optional): Blender frame to start on. Defaults to 1.
        refresh_ui (bool, optional): Refresh the Blender UI on every step. Defaults to False.

    Returns:
        int: step id

    Yields:
        Iterator[int]: Step id
    """
    assert num_steps is not None, "Invalid num_steps"
    assert num_steps > 0, "Invalid num_steps"
    scene = zpy.blender.verify_blender_scene()
    step_idx = 0
    if framerate > 0:
        start = scene.frame_start
        stop = scene.frame_end
        log.info(f"Animation enabled. Min frames: {start}. Max frames: {stop}")
    while step_idx < num_steps:
        zpy.logging.linebreaker_log("step")
        log.info(f"Simulation step {step_idx + 1} of {num_steps}.")
        start_time = time.time()
        if framerate > 0:
            current_frame = start_frame + step_idx * framerate
            scene.frame_set(current_frame)
            log.info(f"Animation frame {scene.frame_current}")
        yield step_idx
        step_idx += 1
        duration = time.time() - start_time
        log.info(f"Simulation step took {duration}s to complete.")
        # TODO: This call is not needed in headless instances, makes loop faster
        if refresh_ui:
            refresh_blender_ui()


def get_asset_lib_path() -> Path:
    """Returns path to asset library location.

    Defaults to the directory of the blenderfile.

    Returns:
        Path: pathlib.Path object to library location.
    """

    assets_env_path = os.environ.get("ASSETS", None)
    if assets_env_path is None:
        log.warning("Could not find environment variable $ASSETS")
        blendfile_path = bpy.path.abspath(bpy.data.filepath)
        return Path(blendfile_path).parent
    else:
        assets_env_path = zpy.files.verify_path(assets_env_path, check_dir=True)
        log.debug(f"Found assets path at {assets_env_path}")
        return assets_env_path


@gin.configurable
def verify_view_layer(
    view_layer_name: str = "View Layer",
) -> bpy.types.ViewLayer:
    """Get and set the view layer in Blender.

    Args:
        view_layer_name (str, optional): Name for View Layer. Defaults to 'View Layer'.

    Returns:
        bpy.types.ViewLayer: View Layer that will be used at runtime.
    """
    scene = zpy.blender.verify_blender_scene()
    view_layer = scene.view_layers.get(view_layer_name, None)
    if view_layer is None:
        log.debug(f"Could not find view layer {view_layer_name}")
        # Default behavior is to use first view layer
        view_layer = scene.view_layers[0]
    log.debug(f"Setting view layer to {view_layer.name}")
    bpy.context.window.view_layer = view_layer
    return view_layer


@gin.configurable
def verify_blender_scene(
    blender_scene_name: str = "Scene",
) -> bpy.types.Scene:
    """Get and set the scene in Blender.

    Args:
        blender_scene_name (str, optional): Name for Scene. Defaults to 'Scene'.

    Returns:
        bpy.types.Scene: Scene that will be used at runtime.
    """
    scene = bpy.data.scenes.get(blender_scene_name, None)
    if scene is None:
        log.debug(f"Could not find scene {blender_scene_name}")
        # Default behavior is to use the first scene
        scene = bpy.data.scenes[0]
    log.debug(f"Setting scene to {scene.name}")
    bpy.context.window.scene = scene
    return scene


def parse_config(
    text_name: str = "config",
) -> None:
    """Parses the gin config text in Blender.

    Args:
        text_name (str, optional): Name of the config text. Defaults to 'config'.
    """
    _text = bpy.data.texts.get(text_name, None)
    if _text is None:
        log.warning(f"Could not find {text_name} in texts.")
        return
    log.info(f"Loading gin config {text_name}")
    gin.enter_interactive_mode()
    with gin.unlock_config():
        gin.parse_config(_text.as_string())
        gin.finalize()


def save_and_revert(_func):
    """Decorator for saving blenderfile before execution, and
        reverting after execution.

    Args:
        _func (callable): function to be decorated.

    Returns:
        [callable]: Wrapped function.
    """

    @wraps(_func)
    def wrapped_func(*args, **kwargs) -> None:
        log.info("Saving the sim.")
        bpy.ops.wm.save_mainfile()
        try:
            _func(*args, **kwargs)
        except Exception as e:
            log.error(f"Executing {_func.__name__} failed with exception {e}")
            raise e
        finally:
            log.info("Reverting sim to previous savepoint.")
            bpy.ops.wm.revert_mainfile()

    return wrapped_func


def load_text_from_file(
    path: Union[Path, str],
    text_name: str = "",
    open_text: bool = False,
) -> None:
    """Load a file into Blender's internal text UI.

    Args:
        path (Union[Path, str]): Filesystem path.
        text_name (str, optional): Name of Blender text to write to.
    """
    path = zpy.files.verify_path(path)
    if bpy.data.texts.get(text_name, None) is None:
        _text = bpy.data.texts.load(str(path), internal=True)
        _text.name = text_name
    else:
        bpy.data.texts[text_name].from_string(path.read_text())
    if open_text:
        for area in bpy.context.screen.areas:
            if area.type == "TEXT_EDITOR":
                area.spaces[0].text = bpy.data.texts[text_name]


def default_script_template_dir() -> Path:
    """Path to the script templates for zpy addon.

    Returns:
        pathlib.Path: Path to script templates for zpy addon.
    """
    script_path = Path(bpy.utils.script_path_user())
    template_dir = script_path / "addons" / "zpy_addon" / "templates"
    return zpy.files.verify_path(template_dir, check_dir=True)


@gin.configurable
def connect_addon(
    name: str = "zpy_addon", addon_dir: Union[Path, str] = "$BLENDERADDONS"
) -> None:
    """Connects a Blender Addon.

    Args:
        name (str, optional): Name of Addon. Defaults to 'zpy_addon'.
        addon_dir (Union[Path, str], optional): Directory of addons. Defaults to '$BLENDERADDONS'.
    """
    log.debug(f"Connecting Addon {name}.")
    path = f"$BLENDERADDONS/{name}/__init__.py"
    path = zpy.files.verify_path(path, make=False)
    bpy.ops.preferences.addon_install(filepath=str(path))
    bpy.ops.preferences.addon_enable(module=name)


@gin.configurable
def connect_debugger_vscode(
    timeout: int = 3,
) -> None:
    """Connects to a VSCode debugger.

    https://github.com/AlansCodeLog/blender-debugger-for-vscode

    Args:
        timeout (int, optional): Seconds to connect before timeout. Defaults to 3.
    """
    if log.getEffectiveLevel() == logging.DEBUG:
        log.debug("Starting VSCode debugger in Blender.")
        connect_addon("blender-debugger-for-vscode")
        bpy.ops.debug.connect_debugger_vscode()
        for sec in range(timeout):
            log.debug(f"You have {timeout - sec} seconds to connect!")
            time.sleep(1)


def save_debug_blenderfile(
    path: Union[Path, str] = None,
) -> None:
    """Saves an intermediate blenderfile for debugging purposes.

    Args:
        path (Union[Path, str], optional): Output directory path.
    """
    if path is None:
        path = zpy.files.default_temp_path() / "_debug.blend"
    path = zpy.files.verify_path(path, make=False)
    log.debug(f"Saving intermediate blenderfile to {path}")
    bpy.ops.wm.save_as_mainfile(filepath=str(path), compress=False, copy=True)


def refresh_blender_ui() -> None:
    """Refresh the Blender UI.

    Does not work on headless instances.
    """
    log.debug("Refreshing Blender UI.")
    bpy.ops.wm.redraw_timer(type="DRAW_WIN_SWAP", iterations=1)
    view_layer = zpy.blender.verify_view_layer()
    view_layer.update()


def load_sim(
    path: Union[Path, str],
    auto_execute_scripts: bool = True,
) -> None:
    """Load a sim from a path to a *.blend file.

    Args:
        path (Union[Path, str]): Path to .blend file.
        auto_execute_scripts (bool, optional): Whether to allow auto execution of scripts. Defaults to True.
    """
    # HACK: Clear out scene of cameras and lights
    clear_scene(["CAMERA", "LIGHT"])
    path = zpy.files.verify_path(path, make=False)
    log.debug(f"Loading sim from {str(path)}.")
    with bpy.data.libraries.load(str(path)) as (data_from, data_to):
        for attr in dir(data_to):
            setattr(data_to, attr, getattr(data_from, attr))
    # HACK: Delete current empty scene
    bpy.ops.scene.delete()
    # HACK: Delete extra workspaces that are created e.g. 'Animation.001'
    _workspaces = [ws for ws in bpy.data.workspaces if ".0" in ws.name]
    bpy.data.batch_remove(ids=_workspaces)
    # Allow execution of scripts inside loaded sim
    if auto_execute_scripts:
        log.warning("Allowing .blend file to run scripts automatically")
        log.warning("   this is unsafe for untrusted files")
        bpy.context.preferences.filepaths.use_scripts_auto_execute = (
            auto_execute_scripts
        )


def clear_scene(
    to_clear: List = ["MESH"],
) -> None:
    """Cleans objects in a scene based on the object type.

    Args:
        to_clear (List, optional): List of object types to clean. Defaults to ["MESH"].
    """
    log.debug(f"Deleting all objects of type {to_clear}")
    for obj in bpy.data.objects:
        if obj.type in to_clear:
            bpy.data.objects.remove(obj)


def scene_information() -> Dict:
    """Returns information on the scene, such as the kwargs in the run text.

    Raises:
        ValueError: Lack of run text and issues with the run text.

    Returns:
        Dict: Sim information dictionary.
    """
    log.info("Collecting scene information")
    run_script = bpy.data.texts.get("run", None)
    if run_script is None:
        raise ValueError("No run script found in scene.")
    # HACK: Gin is confused by the as_module() call
    gin.enter_interactive_mode()
    run_script_module = bpy.data.texts["run"].as_module()
    scene_doc = inspect.getdoc(run_script_module)

    run_function = None
    for name, value in inspect.getmembers(run_script_module):
        if name == "run":
            run_function = value
    if run_function is None:
        raise ValueError("No run() function found in run script.")
    if not inspect.isfunction(run_function):
        raise ValueError("run() is not a function in run script.")

    run_kwargs = []
    for param in inspect.signature(run_function).parameters.values():
        _kwarg = {}
        _kwarg["name"] = param.name
        _kwarg["type"] = str(param.annotation)
        _kwarg["default"] = param.default
        run_kwargs.append(_kwarg)

    scene = zpy.blender.verify_blender_scene()
    _ = {
        "name": scene.zpy_sim_name,
        "version": scene.zpy_sim_version,
        "description": scene_doc,
        "run_kwargs": run_kwargs,
        "export_date": time.strftime("%m%d%Y_%H%M_%S"),
        "zpy_version": zpy.__version__,
        "zpy_path": zpy.__file__,
        "blender_version": ".".join([str(_) for _ in bpy.app.version]),
    }
    log.info(f"{_}")
    return _



"""
    Utilities for Materials in Blender.
"""
import logging
import copy
import random
from pathlib import Path
from typing import Tuple, Union

import bpy

import gin
import zpy

log = logging.getLogger(__name__)


def verify(
    mat: Union[bpy.types.Material, str],
    check_none: bool = True,
) -> bpy.types.Material:
    """Get a material given either its name or the object itself.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        check_none (bool, optional): Check to make sure material exists. Defaults to True.

    Raises:
        ValueError: Material does not exist.

    Returns:
        bpy.types.Material: Material object.
    """
    if isinstance(mat, str):
        mat = bpy.data.materials.get(mat)
    if check_none and mat is None:
        raise ValueError(f"Could not find material {mat}.")
    return mat


def for_mat_in_obj(
    obj: Union[bpy.types.Object, str],
) -> bpy.types.Material:
    """Yield materials in scene object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name)

    Raises:
        ValueError: [description]

    Returns:
        bpy.types.Material: Material object.
    """
    obj = zpy.objects.verify(obj)
    if len(obj.material_slots) > 1:
        for mat in obj.material_slots:
            yield mat.material
    else:
        if obj.active_material is not None:
            return obj.active_material
        else:
            log.debug(f"No active material or material slots found for {obj.name}")
            return None


_SAVED_MATERIALS = {}


def save_mat_props(
    mat: Union[bpy.types.Material, str],
) -> None:
    """Save a pose (rot and pos) to dict.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
    """
    log.info(f"Saving material properties for {mat.name}")
    _SAVED_MATERIALS[mat.name] = get_mat_props(mat)


def restore_mat_props(
    mat: Union[bpy.types.Material, str],
) -> None:
    """Restore an object to a position.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
    """
    log.info(f"Restoring material properties for {mat.name}")
    set_mat_props(mat, _SAVED_MATERIALS[mat.name])


def restore_all_mat_props() -> None:
    """Restore all jittered materials to original look."""
    for mat_name, mat_props in _SAVED_MATERIALS.items():
        set_mat_props(mat_name, mat_props)


def get_mat_props(
    mat: Union[bpy.types.Material, str],
) -> Tuple[float]:
    """Get (some of the) material properties.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)

    Returns:
        Tuple[float]: Material property values (roughness, metallic, specular).
    """
    mat = verify(mat)
    bsdf_node = mat.node_tree.nodes.get("Principled BSDF")
    if bsdf_node is None:
        log.warning(f"No BSDF node in {mat.name}")
        return (0.0, 0.0, 0.0)
    return (
        bsdf_node.inputs["Roughness"].default_value,
        bsdf_node.inputs["Metallic"].default_value,
        bsdf_node.inputs["Specular"].default_value,
    )


def set_mat_props(
    mat: Union[bpy.types.Material, str],
    prop_tuple: Tuple[float],
) -> None:
    """Set (some of the) material properties.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        prop_tuple (Tuple[float]): Material property values (roughness, metallic, specular).
    """
    mat = verify(mat)
    # TODO: Work backwards from Material output node instead of
    #       assuming a 'Principled BSDF' node
    bsdf_node = mat.node_tree.nodes.get("Principled BSDF", None)
    if bsdf_node is None:
        log.warning(f"No BSDF node in {mat.name}")
        return
    bsdf_node.inputs["Roughness"].default_value = copy.copy(prop_tuple[0])
    bsdf_node.inputs["Metallic"].default_value = copy.copy(prop_tuple[1])
    bsdf_node.inputs["Specular"].default_value = copy.copy(prop_tuple[2])


@gin.configurable
def jitter(
    mat: Union[bpy.types.Material, str],
    std: float = 0.2,
    save_first_time: bool = True,
) -> None:
    """Randomize an existing material a little.

    Args:
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        std (float, optional): Standard deviation of gaussian noise over material property. Defaults to 0.2.
        save_first_time (bool, optional): Save the material props first time jitter is called and
            restore before jittering every subsequent time. Defaults to True.
    """
    mat = verify(mat)
    if save_first_time:
        if _SAVED_MATERIALS.get(mat.name, None) is None:
            save_mat_props(mat)
        else:
            restore_mat_props(mat)
    log.info(f"Jittering material {mat.name}")
    mat_props = get_mat_props(mat)
    jittered_mat_props = tuple(map(lambda p: p + random.gauss(0, std), mat_props))
    set_mat_props(mat, jittered_mat_props)


@gin.configurable
def random_texture_mat(
    texture_dir: Union[Path, str] = "lib/textures/random_512p",
    relative_to_assets_dir: bool = True,
) -> bpy.types.Material:
    """Generate a random material from a directory of random texture images.

    Args:
        texture_dir (Union[Path, str], optional): Path to directory with texture images.
        relative_to_assets_dir (bool, optional): Path is relative to the $ASSETS directory. Defaults to False.

    Returns:
        bpy.types.Material: The newly created material.
    """
    if relative_to_assets_dir:
        texture_dir = zpy.blender.get_asset_lib_path().joinpath(texture_dir)
    texture_dir = zpy.files.verify_path(texture_dir, check_dir=True)
    # Create list of texture images in directory
    texture_paths = []
    for _path in texture_dir.iterdir():
        if _path.is_file() and _path.suffix in [".jpg", ".png"]:
            texture_paths.append(_path)
    texture_path = random.choice(texture_paths)
    log.info(f"Found {len(texture_paths)} Textures at {texture_dir}")
    log.info(f"Randomly picked {texture_path.stem}")
    return make_mat_from_texture(texture_path, name=texture_path.stem)


@gin.configurable
def make_mat_from_texture(
    texture_path: Union[Path, str],
    name: str = None,
    coordinate: str = "uv",
) -> bpy.types.Material:
    """Makes a material from a texture image.

    Args:
        texture_path (Union[Path, str]): Path to texture image.
        name (str, optional): Name of new material.
        coordinate (str, optional): Type of texture coordinates. Values are
            "generated", "normal", "uv", "object" , defaults to "uv"

    Returns:
        bpy.types.Material: The newly created material.
    """
    texture_path = zpy.files.verify_path(texture_path, make=False)
    if name is None:
        name = texture_path.stem
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f"Material {name} does not exist, creating it.")
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get("Principled BSDF")
    out_node = mat.node_tree.nodes.get("Material Output")
    tex_node = mat.node_tree.nodes.new("ShaderNodeTexImage")
    tex_node.name = "ImageTexture"
    coord_node = mat.node_tree.nodes.new("ShaderNodeTexCoord")
    bpy.ops.image.open(filepath=str(texture_path))
    tex_node.image = bpy.data.images[texture_path.name]
    tex_node.image.colorspace_settings.name = "Filmic Log"
    mat.node_tree.links.new(tex_node.outputs[0], bsdf_node.inputs[0])
    # TODO: Texture coordinate index is hardcoded
    valid_coordinates = ["generated", "normal", "uv", "object"]
    assert (
        coordinate in valid_coordinates
    ), f"Texture coordinate {coordinate} must be in {valid_coordinates}"
    _coord_idx = valid_coordinates.index(coordinate)
    mat.node_tree.links.new(coord_node.outputs[_coord_idx], tex_node.inputs[0])
    mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    tex_node.image.reload()
    return mat


@gin.configurable
def make_mat_from_color(
    color: Tuple[float],
    name: str = None,
) -> bpy.types.Material:
    """Makes a material given a color.

    Args:
        color (Tuple[float]): Color tuple (RGB).
        name (str, optional): Name of new material.

    Returns:
        bpy.types.Material: The newly created material.
    """
    if name is None:
        name = str(color)
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f"Material {name} does not exist, creating it.")
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get("Principled BSDF")
    out_node = mat.node_tree.nodes.get("Material Output")
    mat.node_tree.nodes.remove(bsdf_node)
    bsdf_node = mat.node_tree.nodes.new("ShaderNodeBsdfDiffuse")
    bsdf_node.inputs["Color"].default_value = color + (1.0,)
    mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    return mat


def set_mat(
    obj: Union[bpy.types.Object, str],
    mat: Union[bpy.types.Material, str],
    recursive: bool = True,
) -> None:
    """Set the material for an object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name) with an active material.
        mat (Union[bpy.types.Material, str]):  Material (or it's name)
        recursive (bool, optional): Recursively set material for child objects. Defaults to True.
    """
    obj = zpy.objects.verify(obj)
    mat = zpy.material.verify(mat)
    if hasattr(obj, "active_material"):
        log.debug(f"Setting object {obj.name} material {mat.name}")
        obj.active_material = mat
    else:
        log.warning("Object does not have material property")
        return
    # Recursively change material on all children of object
    if recursive:
        for child in obj.children:
            set_mat(child, mat)


@gin.configurable
def make_aov_material_output_node(
    mat: bpy.types.Material = None,
    obj: bpy.types.Object = None,
    style: str = "instance",
) -> None:
    """Make AOV Output nodes in Composition Graph.

    Args:
        mat (bpy.types.Material, optional): A blender material (either it's name or the object itself).
        obj (bpy.types.Object, optional): A blender object (either it's name or the object itself).
        style (str, optional): Type of segmentation in [instance, category]. Defaults to 'instance'.

    Raises:
        ValueError: Invalid style, no object or material given.
    """
    # Make sure engine is set to Cycles
    scene = zpy.blender.verify_blender_scene()
    if not (scene.render.engine == "CYCLES"):
        log.warning(" Setting render engine to CYCLES to use AOV")
        scene.render.engine == "CYCLES"

    # TODO: Refactor this legacy "styles" code

    # Only certain styles are available
    valid_styles = ["instance", "category"]
    assert (
        style in valid_styles
    ), f"Invalid style {style} for AOV material output node, must be in {valid_styles}."

    # HACK: multiple material slots
    all_mats = []

    # Use material
    if mat is not None:
        all_mats = [mat]
    # Get material from object
    elif obj is not None:
        if obj.active_material is None:
            log.debug(f"No active material found for {obj.name}")
            return
        if len(obj.material_slots) > 1:
            for mat in obj.material_slots:
                all_mats.append(mat.material)
        else:
            all_mats.append(obj.active_material)
    else:
        raise ValueError("Must pass in an Object or Material")

    # HACK: multiple material slots
    for mat in all_mats:

        # Make sure material is using nodes
        if not mat.use_nodes:
            mat.use_nodes = True
        tree = mat.node_tree

        # Vertex Color Node
        vcol_node = zpy.nodes.get_or_make(
            f"{style} Vertex Color", "ShaderNodeVertexColor", tree
        )
        vcol_node.layer_name = style

        # AOV Output Node
        # HACK: This type of node has a "name" property which prevents using the
        # normal zpy.nodes code due to a scope conflict with the bpy.types.Node.name property
        # See: https://docs.blender.org/api/current/bpy.types.ShaderNodeOutputAOV.html
        _name = style
        aovout_node = None
        for _node in tree.nodes:
            if _node.name == _name:
                aovout_node = _node
        if aovout_node is None:
            aovout_node = tree.nodes.new("ShaderNodeOutputAOV")
        aovout_node.name = style

        tree.links.new(vcol_node.outputs["Color"], aovout_node.inputs["Color"])



#!/usr/bin/env python
from distutils.core import setup
from setuptools import find_packages
import versioneer


def get_requirements_from_file(python_requirements_file="./requirements.txt"):
    """
    Purpose:
        Get python requirements from a specified requirements file.
    Args:
        python_requirements_file (String): Path to the requirements file (usually
            it is requirements.txt in the same directory as the setup.py)
    Return:
        requirements (List of Strings): The python requirements necessary to run
            the library
    """

    requirements = []
    with open(python_requirements_file) as requirements_file:
        requirement = requirements_file.readline()
        while requirement:
            if requirement.strip().startswith("#"):
                pass
            elif requirement.strip() == "":
                pass
            else:
                requirements.append(requirement.strip())
            requirement = requirements_file.readline()

    return requirements


def get_long_description(filepath):
    with open(filepath, "r", encoding="utf-8") as fh:
        long_description = fh.read()
    return long_description


setup(
    name="zpy-zumo",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Create synthetic data with Blender.",
    long_description=get_long_description("README.md"),
    long_description_content_type="text/markdown",
    author="Zumo Labs",
    author_email="infra@zumolabs.ai",
    url="https://github.com/ZumoLabs/zpy",
    packages=find_packages(),
    install_requires=get_requirements_from_file(),
    include_package_data=True,
    entry_points="[console_scripts]\nzpy=cli.cli:cli\n",
)



"""
    Export panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """Properties applied to scenes."""
    bpy.types.Scene.zpy_sim_name = bpy.props.StringProperty(
        name="Sim Name",
        description="Name of the scene, must match data portal.",
        default="default",
    )
    bpy.types.Scene.zpy_sim_version = bpy.props.StringProperty(
        name="Sim Version",
        description="Version of the scene, must match data portal.",
        default="0",
    )
    bpy.types.Scene.zpy_export_dir = bpy.props.StringProperty(
        name="Export Directory Path",
        description="Path to directory for exporting packaged zumo scenes.",
        default=str(zpy.files.default_temp_path()),
        subtype="DIR_PATH",
    )
    bpy.types.Scene.zpy_export_path = bpy.props.StringProperty(
        name="Export Path",
        description="Export path for this zumo scene.",
        default=str(zpy.files.default_temp_path()),
        subtype="DIR_PATH",
    )


class OpenExportDirOperator(bpy.types.Operator):
    """Open file browser at export dir."""

    bl_idname = "scene.zpy_open_export_dir"
    bl_label = "Open Export Dir"
    bl_description = "Open file browser at export dir."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        zpy.files.open_folder_in_explorer(context.scene.zpy_export_dir, make=True)
        return {"FINISHED"}


class CleanUpDirOperator(bpy.types.Operator):
    """Clean up and package sim into export dir."""

    bl_idname = "scene.zpy_cleanup_sim"
    bl_label = "Clean Up Export Dir"
    bl_description = "Clean up export dir."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        log.info("Cleaning up scene.")
        # Remove any backup blender files
        zpy.files.remove_files_with_suffix(
            path=context.scene.zpy_export_path,
            exts=[".blend1", ".blend2", ".blend3"],
        )
        # TODO: Scene based clean up collections and objects listings (in the text editor)
        # TODO: Remove the custom scene scripts that are not needed for staging (keep run, config, categories for now)
        return {"FINISHED"}


class ExportOperator(bpy.types.Operator):
    """Export sim for ingest to Data Portal."""

    bl_idname = "scene.zpy_export_sim"
    bl_label = "Export sim"
    bl_description = "Export sim for ingest to Data Portal."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        # Shows progress cursor in Blender UI.
        log.info("Export Started.")
        bpy.context.window_manager.progress_begin(0, 100)

        # Clean scene before every export
        bpy.ops.scene.zpy_cleanup_sim()

        # Create export directory in the Blender filepath
        export_dir_name = (
            f"{context.scene.zpy_sim_name}_v{context.scene.zpy_sim_version}"
        )
        export_path = Path(context.scene.zpy_export_dir) / export_dir_name
        zpy.files.verify_path(export_path, make=True)

        # Set the scene export path
        context.scene.zpy_export_path = str(export_path)

        # Find missing files before export
        log.info("Export Step 1 of 4: Checking for any missing files.")
        bpy.context.window_manager.progress_update(10)
        _path = zpy.files.verify_path("$ASSETS", make=False)
        bpy.ops.file.find_missing_files(directory=str(_path))

        # Fix all the asset paths by packing them into the .blend
        # file and then un-packing them into a 'textures' folder.
        try:
            bpy.ops.wm.save_as_mainfile(
                filepath=str(export_path / "main.blend"),
                compress=False,
                relative_remap=True,
            )
            log.info("Export Step 2 of 4: Packing files into .blend.")
            bpy.context.window_manager.progress_update(30)
            bpy.ops.file.make_paths_absolute()
            bpy.ops.file.make_paths_relative()
            bpy.ops.file.pack_all()
            bpy.ops.file.unpack_all(method="WRITE_LOCAL")
            bpy.ops.wm.save_as_mainfile(
                filepath=str(export_path / "main.blend"),
                compress=False,
                relative_remap=True,
            )
        except Exception as e:
            self.report({"ERROR"}, str(e))
            log.warning(f"Exception when exporting: {e}")
            return {"CANCELLED"}

        log.info("Export Step 3 of 4: Saving meta-information.")
        bpy.context.window_manager.progress_update(70)

        # Output scene information in _meta.zumo
        zpy.files.write_json(
            export_path / "_meta.zumo.json",
            zpy.blender.scene_information(),
        )

        # TODO: Export glTF into zip directory

        # Clean up scene before zipping up
        bpy.ops.scene.zpy_cleanup_sim()

        # Zip up the exported directory for easy upload
        log.info("Export Step 4 of 4: Zipping up package.")
        bpy.context.window_manager.progress_update(90)
        zpy.files.zip_file(
            in_path=export_path,
            zip_path=Path(context.scene.zpy_export_dir) / f"{export_dir_name}.zip",
        )

        log.info("Export Completed.")
        bpy.context.window_manager.progress_end()
        return {"FINISHED"}


class SCENE_PT_ExportPanel(bpy.types.Panel):
    """UI for the addon that is visible in Blender."""

    bl_idname = "SCENE_PT_ExportPanel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Export"
    bl_category = "ZPY"
    bl_options = {"DEFAULT_CLOSED"}

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row()
        row.operator(
            "scene.zpy_export_sim",
            text="Export Sim",
            icon="EXPORT",
        )
        row = layout.row()
        row.prop(scene, "zpy_sim_name", text="Name")
        row = layout.row()
        row.prop(scene, "zpy_sim_version", text="Version")
        row = layout.row()
        row.label(text="Export Path")
        row = layout.row()
        row.prop(scene, "zpy_export_dir", text="")
        row = layout.row()
        row.operator(
            "scene.zpy_open_export_dir",
            text="Open",
            icon="FILEBROWSER",
        )
        row.operator(
            "scene.zpy_cleanup_sim",
            text="Clean",
            icon="PACKAGE",
        )



from zpy.files import to_pathlib_path
import os
import yaml

ENDPOINTS = {
    "local": "http://localhost:8000",
    "stage": "https://ragnarok.stage.zumok8s.org",
    "prod": "https://ragnarok.zumok8s.org",
}
CONFIG_FILE = "~/.zpy/config.yaml"


def get_endpoint(env):
    """get endpoint

    Given an env return endpoint
    Args:
        env (str): desired env

    Returns:
        url (str): endpoint for env
    """
    return ENDPOINTS[env]


def initialize_config():
    """initialize config

    If CONFIG_FILE doesnt exist write it and put in prod as the endpoint. Also creates
    the ~/.zpy folder if not existing. The config is some variables needed by the cli to
    make validated requests to the backend.
    """
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    if path.exists():
        return
    CONFIG = {"ENVIRONMENT": "prod", "TOKEN": None, "ENDPOINT": ENDPOINTS["prod"]}
    path.parent.mkdir(parents=True, exist_ok=True)
    write_config(CONFIG)


def read_config():
    """read config

    Read zpy cli configuration file.

    Returns:
        config: dictionary of current configuration
    """
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    with path.open() as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config


def write_config(config):
    """write config

    Write zpy cli configuration file.

    Args:
        config (dict): new configuration to write
    """
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    with path.open("w") as f:
        yaml.dump(config, f)



"""
    Human Keypoint Skeleton Formats.
"""
import logging

import bpy
import gin

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class Keypoints:
    """Functionality for Keypoints."""

    # COCO(From pycoco)
    COCO_NAMES = [
        "nose",
        "left_eye",
        "right_eye",
        "left_ear",
        "right_ear",
        "left_shoulder",
        "right_shoulder",
        "left_elbow",
        "right_elbow",
        "left_wrist",
        "right_wrist",
        "left_hip",
        "right_hip",
        "left_knee",
        "right_knee",
        "left_ankle",
        "right_ankle",
    ]
    COCO_CONNECTIVITY = [
        [16, 14],
        [14, 12],
        [17, 15],
        [15, 13],
        [12, 13],
        [6, 12],
        [7, 13],
        [6, 7],
        [6, 8],
        [7, 9],
        [8, 10],
        [9, 11],
        [2, 3],
        [1, 2],
        [1, 3],
        [2, 4],
        [3, 5],
        [4, 6],
        [5, 7],
    ]
    COCO_BONE_LOOKUP_MIXAMO = {
        "nose": "NoseEnd",
        "left_eye": "LeftEyeEnd",
        "right_eye": "RightEyeEnd",
        "left_ear": "LeftEarEnd",
        "right_ear": "RightEarEnd",
        "left_shoulder": "LeftArm",
        "right_shoulder": "RightArm",
        "left_elbow": "LeftForeArm",
        "right_elbow": "RightForeArm",
        "left_wrist": "LeftHand",
        "right_wrist": "RightHand",
        "left_hip": "LeftUpLeg",
        "right_hip": "RightUpLeg",
        "left_knee": "LeftLeg",
        "right_knee": "RightLeg",
        "left_ankle": "LeftFoot",
        "right_ankle": "RightFoot",
    }
    COCO_BONE_LOOKUP_ANIMA = {
        # TODO: Anima armature does not have eyes, ears, nose
        "nose": "Head",
        "left_eye": "Head",
        "right_eye": "Head",
        "left_ear": "Head",
        "right_ear": "Head",
        "left_shoulder": "LeftArm",
        "right_shoulder": "RightArm",
        "left_elbow": "LeftForeArm",
        "right_elbow": "RightForeArm",
        "left_wrist": "LeftHand",
        "right_wrist": "RightHand",
        "left_hip": "LeftUpLeg",
        "right_hip": "RightUpLeg",
        "left_knee": "LeftLeg",
        "right_knee": "RightLeg",
        "left_ankle": "LeftFoot",
        "right_ankle": "RightFoot",
    }

    # Body25B
    BODY25B_NAMES = [
        "Nose",
        "Neck",
        "RShoulder",
        "RElbow",
        "RWrist",
        "LShoulder",
        "LElbow",
        "LWrist",
        "RHip",
        "RKnee",
        "RAnkle",
        "LHip",
        "LKnee",
        "LAnkle",
        "REye",
        "LEye",
        "REar",
        "LEar",
    ]
    BODY25B_CONNECTIVITY = [
        [0, 1],
        [0, 2],
        [1, 3],
        [2, 4],
        [5, 7],
        [6, 8],
        [7, 9],
        [8, 10],
        [5, 11],
        [6, 12],
        [11, 13],
        [12, 14],
        [13, 15],
        [14, 16],
        [15, 19],
        [19, 20],
        [15, 21],
        [16, 22],
        [22, 23],
        [16, 24],
        [5, 17],
        [6, 17],
        [17, 18],
        [11, 12],
    ]
    BODY25B_BONE_LOOKUP_MIXAMO = {
        "Nose": "NoseEnd",
        "Neck": "Head",
        "RShoulder": "RightShoulder",
        "RElbow": "RightForeArm",
        "RWrist": "RightHand",
        "LShoulder": "LeftShoulder",
        "LElbow": "LeftForeArm",
        "LWrist": "LeftHand",
        "RHip": "RightUpLeg",
        "RKnee": "RightLeg",
        "RAnkle": "RightFoot",
        "LHip": "LeftUpLeg",
        "LKnee": "LeftLeg",
        "LAnkle": "LeftFoot",
        "REye": "RightEyeEnd",
        "LEye": "LeftEyeEnd",
        "REar": "RightEarEnd",
        "LEar": "LeftEarEnd",
    }
    # TODO: Anima armature
    BODY25B_BONE_LOOKUP_ANIMA = {
        "Nose": "Head",
        "Neck": "Head",
        "RShoulder": "RightShoulder",
        "RElbow": "RightForeArm",
        "RWrist": "RightHand",
        "LShoulder": "LeftShoulder",
        "LElbow": "LeftForeArm",
        "LWrist": "LeftHand",
        "RHip": "RightUpLeg",
        "RKnee": "RightLeg",
        "RAnkle": "RightFoot",
        "LHip": "LeftUpLeg",
        "LKnee": "LeftLeg",
        "LAnkle": "LeftFoot",
        "REye": "RightEyeEnd",
        "LEye": "LeftEyeEnd",
        "REar": "RightEarEnd",
        "LEar": "LeftEarEnd",
    }

    def __init__(
        self,
        root: bpy.types.Object,
        style: str = "coco",
        armature: str = "anima",
    ):
        """Initialize keypoint object."""
        if style == "coco":
            self.names = self.COCO_NAMES
            self.connectivity = self.COCO_CONNECTIVITY
            if armature == "mixamo":
                self.bone_lookup = self.COCO_BONE_LOOKUP_MIXAMO
            elif armature == "anima":
                self.bone_lookup = self.COCO_BONE_LOOKUP_ANIMA
            else:
                raise ValueError(f"Unknown keypoint armature: {armature}")
        elif style == "body25b":
            self.names = self.BODY25B_NAMES
            self.connectivity = self.BODY25B_CONNECTIVITY
            if armature == "mixamo":
                self.bone_lookup = self.BODY25B_BONE_LOOKUP_MIXAMO
            elif armature == "anima":
                self.bone_lookup = self.BODY25B_BONE_LOOKUP_ANIMA
            else:
                raise ValueError(f"Unknown keypoint armature: {armature}")
        else:
            raise ValueError(f"Unknown keypoint style: {style}")
        self.root = root
        self.style = style
        self.armature = armature
        self.bones = {bone.name: bone for bone in self.root.pose.bones}
        self.num_keypoints = None
        self.keypoints_xyv = None
        self.keypoints_xyz = None

    def update(
        self,
        world_transform=None,
    ) -> None:
        """Add a keypoint skeleton."""
        self.num_keypoints = 0
        self.keypoints_xyv = []
        self.keypoints_xyz = []
        for name, bone_name in self.bone_lookup.items():
            bone = self.bones.get(bone_name, None)
            if bone is None:
                log.warning(f"Could not find keypoint bone {name} using {bone_name}")
            if world_transform is None:
                pos = self.root.matrix_world @ bone.head
            else:
                pos = world_transform @ self.root.matrix_world @ bone.head
            x, y, v = zpy.camera.camera_xyv(pos, obj=self.root)
            self.keypoints_xyv += [x, y, v]
            self.keypoints_xyz += tuple(pos)
            self.num_keypoints += 1



"""
    Rendering panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy
import gin
from bpy.types import Operator

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


class RunOperator(Operator):
    """Launch the run script in Blender's texts."""

    bl_idname = "scene.zpy_run"
    bl_label = "Run Sim"
    bl_description = "Launch the run script in Blender's texts."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    def execute(self, context):
        text = bpy.data.texts.get("run", None)
        if text is None:
            raise ValueError(
                'Running a sim requires a run text, could not find in text with name "run".'
            )
        # Set the logger levels
        zpy.logging.set_log_levels("debug")
        # HACK: Gin will complain when this module is re-initialized
        with gin.unlock_config():
            text_as_module = text.as_module()
        # Execute the run function inside the run text
        text_as_module.run()
        return {"FINISHED"}


class RenderOperator(Operator):
    """Render out single image (rgb, segmented, depth)."""

    bl_idname = "scene.zpy_render"
    bl_label = "Render Frame"
    bl_description = "Render out segmented images."
    bl_category = "ZPY"
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        # TODO: Make sure sim is good to render?
        return True

    def execute(self, context):
        context.space_data.shading.color_type = "OBJECT"

        # Image names
        rgb_image_name = zpy.files.make_rgb_image_name(0)
        cseg_image_name = zpy.files.make_cseg_image_name(0)
        iseg_image_name = zpy.files.make_iseg_image_name(0)
        depth_image_name = zpy.files.make_depth_image_name(0)

        # Output path
        output_path = Path(context.scene.zpy_output_path)

        # Save renders to file
        zpy.render.render(
            rgb_path=output_path / rgb_image_name,
            iseg_path=output_path / iseg_image_name,
            cseg_path=output_path / cseg_image_name,
            depth_path=output_path / depth_image_name,
            width=640,
            height=480,
        )

        return {"FINISHED"}


class SCENE_PT_ExecutePanel(bpy.types.Panel):
    """UI for the addon that is visible in Blender."""

    bl_idname = "SCENE_PT_ExecutePanel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Execute"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout

        row = layout.row()
        row.operator(
            "scene.zpy_render",
            text="Render",
            icon="FILE_IMAGE",
        )
        row = layout.row()
        row.operator(
            "scene.zpy_run",
            text="Run (Debug)",
            icon="TRIA_RIGHT",
        )



"""
    Camera utilities.
"""
import logging
import math
from typing import Tuple, Union

import bpy
import bpy_extras
import gin
import mathutils

import zpy

log = logging.getLogger(__name__)


def verify(
    camera: Union[bpy.types.Object, bpy.types.Camera, str],
    check_none=True,
) -> bpy.types.Camera:
    """Return camera given name or typed object.

    Args:
        camera (Union[bpy.types.Object, bpy.types.Camera, str]): Camera object (or it's name)
        check_none (bool, optional): Raise error if object does not exist. Defaults to True.

    Raises:
        ValueError: Object does not exist.

    Returns:
        bpy.types.Camera: Camera object.
    """
    if isinstance(camera, str):
        camera = bpy.data.cameras.get(camera)
    if check_none and camera is None:
        raise ValueError(f"Could not find camera {camera}.")
    if camera is None:
        log.info(f'No camera chosen, using default scene camera "{camera}".')
        scene = zpy.blender.verify_blender_scene()
        camera = scene.camera
    return camera


def look_at(
    obj: Union[bpy.types.Object, str],
    location: Union[Tuple[float], mathutils.Vector],
    roll: float = 0,
) -> None:
    """Rotate obj to look at target.

    Based on: https://blender.stackexchange.com/a/5220/12947

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name) that does the looking (usually a camera)
        location (Union[Tuple[float], mathutils.Vector]): Location (3-tuple or Vector) to be looked at.
        roll (float, optional): The angle of rotation about the axis from obj to target in radians. Defaults to 0.
    """
    obj = zpy.objects.verify(obj)
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    loc = obj.location
    # direction points from the object to the target
    direction = location - obj.location
    quat = direction.to_track_quat("-Z", "Y")
    quat = quat.to_matrix().to_4x4()
    # convert roll from radians to degrees
    roll_matrix = mathutils.Matrix.Rotation(roll, 4, "Z")
    # remember the current location, since assigning to obj.matrix_world changes it
    loc = loc.to_tuple()
    obj.matrix_world = quat @ roll_matrix
    obj.location = loc


@gin.configurable
def camera_xyz(
    location: Union[Tuple[float], mathutils.Vector],
    camera: Union[bpy.types.Object, bpy.types.Camera, str] = None,
    fisheye_lens: bool = False,
) -> Tuple[float]:
    """Get pixel coordinates of point in camera space.

    - (0, 0) is the bottom left of the camera frame.
    - (1, 1) is the top right of the camera frame.
    - Values outside 0-1 are also supported.
    - A negative ‘z’ value means the point is behind the camera.

    Args:
        location (mathutils.Vector): Location (3-tuple or Vector) of point in 3D space.
        camera (Union[bpy.types.Object, bpy.types.Camera, str]): Camera in which pixel space exists.
        fisheye_lens (bool, optional): Whether to use fisheye distortion. Defaults to False.

    Returns:
        Tuple[float]: Pixel coordinates of location.
    """
    camera = zpy.camera.verify(camera)
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    scene = zpy.blender.verify_blender_scene()
    point = bpy_extras.object_utils.world_to_camera_view(scene, camera, location)
    # TODO: The z point here is incorrect?
    log.debug(f"Point {point}")
    if point[2] < 0:
        log.debug("Point is behind camera")

    # Fix the point based on camera distortion
    if fisheye_lens:
        log.debug("Correcting for fisheye distortion")

        # HACK: There should be a better place to put this
        bpy.data.cameras[0].lens_unit = "FOV"
        bpy.data.cameras[0].lens = 18.0

        # https://blender.stackexchange.com/questions/40702/how-can-i-get-the-projection-matrix-of-a-panoramic-camera-with-a-fisheye-equisol?noredirect=1&lq=1
        # Note this assumes 180 degree FOV
        cam = bpy.data.cameras[camera.name]
        f = cam.lens
        w = cam.sensor_width
        h = cam.sensor_height

        p = camera.matrix_world.inverted() @ location
        p.normalize()

        # Calculate our angles
        phi = math.atan2(p.y, p.x)
        length = (p.x ** 2 + p.y ** 2) ** (1 / 2)
        theta = math.asin(length)

        # Equisolid projection
        r = 2.0 * f * math.sin(theta / 2)

        u = r * math.cos(phi) / w + 0.5
        v = r * math.sin(phi) / h + 0.5

        # x = u * scene.render.resolution_x
        # y = v * scene.render.resolution_y
        # TODO: The value of point[2] here is not exactly correct ...
        return u, v, point[2]

    else:
        return point[0], point[1], point[2]


def is_child_hit(
    obj: Union[bpy.types.Object, str],
    hit_obj: Union[bpy.types.Object, str],
) -> bool:
    """Recursive function to check if a child object is the hit object.

    Args:
        obj (Union[bpy.types.Object, str]): Scene object (or it's name) that might contain a hit child.
        hit_obj (Union[bpy.types.Object, str]): Scene object (or it's name) that was hit

    Returns:
        bool: If the hit object is a child object.
    """
    obj = zpy.objects.verify(obj)
    hit_obj = zpy.objects.verify(hit_obj)
    if obj == hit_obj:
        return True
    else:
        for child in obj.children:
            if is_child_hit(child, hit_obj):
                return True
        return False


def is_visible(
    location: Union[Tuple[float], mathutils.Vector],
    obj_to_hit: Union[bpy.types.Object, str],
    camera: Union[bpy.types.Object, bpy.types.Camera, str] = None,
) -> bool:
    """Cast a ray to determine if object is visible from camera.

    Args:
        location (Union[Tuple[float], mathutils.Vector]): Location to shoot out ray towards.
        obj_to_hit (Union[bpy.types.Object, str]): Object that should be hit by ray.
        camera (Union[bpy.types.Object, bpy.types.Camera, str]): Camera where ray originates from.

    Returns:
        bool: Whether the casted ray has hit the object.
    """
    camera = zpy.camera.verify(camera)
    obj_to_hit = zpy.objects.verify(obj_to_hit)
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    view_layer = zpy.blender.verify_view_layer()
    scene = zpy.blender.verify_blender_scene()
    result = scene.ray_cast(
        depsgraph=view_layer.depsgraph,
        origin=camera.location,
        direction=(location - camera.location),
    )
    # Whether a hit occured
    is_hit = result[0]
    # Object hit by raycast
    hit_obj = result[4]
    if not is_hit:
        # Nothing was hit by the ray
        log.debug(f"No raycast hit from camera to {obj_to_hit.name}")
        return False
    if is_child_hit(obj_to_hit, hit_obj):
        # One of the children of the obj_to_hit was hit
        log.debug(f"Raycast hit from camera to {obj_to_hit.name}")
        return True
    return False


@gin.configurable
def is_in_view(
    location: Union[Tuple[float], mathutils.Vector],
    camera: Union[bpy.types.Object, bpy.types.Camera, str] = None,
    epsilon: float = 0.05,
) -> bool:
    """Is a location visible from a camera (within some epsilon).

    Args:
        location (Union[Tuple[float], mathutils.Vector]): Location that is visible or not.
        camera (Union[bpy.types.Object, bpy.types.Camera, str]): Camera that wants to see the location.
        epsilon (float, optional): How far outside the view box the point is allowed to be. Defaults to 0.05.

    Returns:
        bool: Whether the location is visible.
    """
    camera = zpy.camera.verify(camera)
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    x, y, z = camera_xyz(location, camera=camera)
    if z < 0:
        return False
    if x < (0 - epsilon) or x > (1 + epsilon):
        return False
    if y < (0 - epsilon) or y > (1 + epsilon):
        return False
    return True


@gin.configurable
def camera_xyv(
    location: Union[Tuple[float], mathutils.Vector],
    obj: Union[bpy.types.Object, str],
    camera: Union[bpy.types.Object, bpy.types.Camera, str] = None,
    width: int = 640,
    height: int = 480,
) -> Tuple[int]:
    """Get camera image xyv coordinates of point in scene.

    Keypoint coordinates (x, y) are measured from the top left
    image corner (and are 0-indexed). Coordinates are rounded
    to the nearest pixel to reduce file size. Visibility (v):

        v=0: not labeled (in which case x=y=0)
        v=1: labeled but not visible
        v=2: labeled and visible

    Args:
        location (Union[Tuple[float], mathutils.Vector]): Location (3-tuple or Vector) of point in 3D space.
        obj (Union[bpy.types.Object, str]): Scene object (or it's name) to check for visibility.
        camera (Union[bpy.types.Object, bpy.types.Camera, str]): Camera in which pixel space exists.
        width (int, optional): Width of image. Defaults to 640.
        height (int, optional): Height of image. Defaults to 480.

    Returns:
        Tuple[int]: (X, Y, V)
    """
    camera = zpy.camera.verify(camera)
    obj = zpy.objects.verify(obj)
    if not isinstance(location, mathutils.Vector):
        location = mathutils.Vector(location)
    x, y, z = camera_xyz(location, camera=camera)
    # visibility
    v = 2
    if x < 0 or y < 0 or z < 0:
        v = 1
    if not is_visible(location, obj_to_hit=obj, camera=camera):
        v = 1
    # bottom-left to top-left
    y = 1 - y
    # float (0, 1) to pixel int (0, pixel size)
    x = int(x * width)
    y = int(y * height)
    log.debug(f"(x, y, v) {(x, y, v)}")
    return x, y, v



from cli.utils import fetch_auth, download_url
from zpy.files import to_pathlib_path
import json
import requests


@fetch_auth
def create_sim(name, path, url, auth_headers):
    """create sim

    Upload sim object to S3 through ZumoLabs backend and create
    the sim object.

    Args:
        name (str): name of sim to upload
        path (str): file to upload
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/sims/"
    r = requests.post(
        endpoint,
        data={"name": name},
        files={"file": open(path, "rb")},
        headers=auth_headers,
    )
    if r.status_code != 201:
        r.raise_for_status()


@fetch_auth
def download_sim(name, path, url, auth_headers):
    """download sim

    Download sim object from S3 through ZumoLabs backend.

    Args:
        name (str): name of sim to download
        path (str): output directory
        url (str): backend endpoint
        auth_headers: authentication for backend

    Returns:
        str: output file path
    """
    endpoint = f"{url}/api/v1/sims/"
    r = requests.get(endpoint, params={"name": name}, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    if response["count"] != 1:
        raise NameError(f"found {response['count']} sims for name {name}")
    sim = response["results"][0]
    endpoint = f"{url}/api/v1/sims/{sim['id']}/download"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    response = json.loads(r.text)
    name_slug = f"{sim['name'].replace(' ', '_')}-{sim['id'][:8]}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(response["redirect_link"], output_path)
    return output_path


@fetch_auth
def fetch_sims(url, auth_headers):
    """fetch sims

    Fetch sim objects from ZumoLabs backend.

    Args:
        url (str): backend endpoint
        auth_headers: authentication for backend
    """
    endpoint = f"{url}/api/v1/sims/"
    r = requests.get(endpoint, headers=auth_headers)
    if r.status_code != 200:
        r.raise_for_status()
    return json.loads(r.text)["results"]
